/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "editorwidget.h"
#include "synhighlighter.h"

#include <QApplication>
#include <QMessageBox>
#include <QPainter>
#include <QTextBlock>
#include <QTextStream>
#include <Qt>

TextEditorWidget::TextEditorWidget() {
    setLineWrapMode(QPlainTextEdit::NoWrap);

    // font
    QFont font;
    font.setFamily("monospace");
    font.setStyleHint(QFont::Monospace);
    font.setFixedPitch(true);
    font.setPointSize(12);
    QFontMetrics metrics(font);
    // TODO: tab width configurable
    setTabStopWidth(metrics.horizontalAdvance('0') * 4);
    setFont(font);

    connect(this, SIGNAL(updateRequest(QRect, int)), this, SLOT(updateInfoStrip(QRect, int)));
    connect(document(), SIGNAL(modificationChanged(bool)), this, SLOT(contentsModified(bool)));

    // info-strip for line numbers + markers
    infoStrip = new InfoStrip(this);
    setViewportMargins(infoStrip->stripWidth(), 0, 0, 0);

    // syntax highlighter
    new SynHighlighter(document());
}

bool TextEditor::load() {
    QFile file(path());
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
        warning(mwidget, QString("Error opening %1:\n%2").arg(path()).arg(file.errorString()));
        return false;
    }

    QTextStream in(&file);
    QApplication::setOverrideCursor(Qt::WaitCursor);
    mwidget->setPlainText(in.readAll());
    QApplication::restoreOverrideCursor();
    return true;
}

bool TextEditor::save() {
    QFile file(path());
    if (!file.open(QFile::WriteOnly | QFile::Text)) {
        warning(mwidget, QString("Error saving %1:\n%2.").arg(path()).arg(file.errorString()));
        return false;
    }
    QTextStream out(&file);
    QApplication::setOverrideCursor(Qt::WaitCursor);
    out << mwidget->toPlainText();
    QApplication::restoreOverrideCursor();
    mwidget->document()->setModified(false);
    return true;
}

QList<EditorLink> TextEditor::findAll(const SearchTerm &term, bool active) {
    QList<EditorLink> ret;
    // grab current cursor location
    QTextCursor cursor = mwidget->textCursor();
    // find and hilite all results
    mwidget->moveCursor(QTextCursor::Start);
    QList<QTextEdit::ExtraSelection> selections;
    QTextDocument::FindFlags options;
    bool sensitive = term.caseSensitive();
    options.setFlag(QTextDocument::FindFlag::FindCaseSensitively, sensitive);
    while (mwidget->find(term.string(), options) && ret.length() <= 500) {
        QTextCursor cursor = mwidget->textCursor();
        QTextEdit::ExtraSelection extra;
        extra.cursor = cursor;
        extra.format.fontUnderline();
        QColor color(222, 247, 177);
        extra.format.setBackground(color);
        selections.append(extra);
        int anchor = cursor.anchor();
        // expand selection to grab more text
        cursor.movePosition(QTextCursor::EndOfWord, QTextCursor::KeepAnchor, 12);
        // find snippet
        QTextBlock block = mwidget->document()->findBlock(anchor);
        QString line = block.text(); // TODO: is this correct
        int index = anchor - block.position();
        int space = line.indexOf(QRegExp("\\s+"), index + term.string().size() + 1);
        int end = (space == -1 ? line.size() : space);
        QString text = line.mid(index, end - index);
        int position = cursor.position();
        EditorLink link(text, file(), anchor, position);
        ret.append(link);
    }
    mwidget->setExtraSelections(selections);
    // return to original cursor
    mwidget->setTextCursor(cursor);
    // if active editor find next
    if (active) {
        if (!cursor.selectedText().endsWith(term.string(),
                                            sensitive ? Qt::CaseSensitive : Qt::CaseInsensitive)) {
            findNext(term);
        }
    }
    return ret;
}

void TextEditor::findNext(const SearchTerm &term) {
    QTextDocument::FindFlags options;
    options.setFlag(QTextDocument::FindFlag::FindCaseSensitively, term.caseSensitive());
    if (!mwidget->find(term.string(), options)) {
        mwidget->moveCursor(QTextCursor::Start);
        if (!mwidget->find(term.string(), options)) {
            warning(mwidget, "'" + term.string() + "' not found in current file");
        }
    }
}

void TextEditor::replaceNext(const QString &term, const QString &replace) {
    QString s = mwidget->textCursor().selectedText();
    if (term.compare(s) == 0) { // make sure term is same as selected text
        mwidget->textCursor().insertText(replace);
        mwidget->find(term); // select next
    }
}

void TextEditor::focusLine(int line, int column) {
    QTextCursor cursor = mwidget->textCursor();
    cursor.movePosition(QTextCursor::Start);
    cursor.movePosition(QTextCursor::Down, QTextCursor::MoveAnchor, line - 1);
    if (column >= mwidget->document()->findBlock(line).length()) {
        cursor.movePosition(QTextCursor::EndOfLine);
    } else {
        cursor.movePosition(QTextCursor::Right, QTextCursor::MoveAnchor, column - 1);
    }
    mwidget->setTextCursor(cursor);
    mwidget->setFocus();
}

void TextEditor::focusLocation(EditorLocation &loc) {
    QTextCursor cursor = mwidget->textCursor();
    cursor.setPosition(loc.position());
    cursor.setPosition(loc.anchor(), QTextCursor::KeepAnchor);
    mwidget->setTextCursor(cursor);
    mwidget->setFocus();
}

void TextEditor::commentSelection() {
    Selection sel = mwidget->currentSelection();
    // uncomment if all lines currently commented
    bool comment = false;
    for (int i = sel.start; i <= sel.end; i++) {
        if (!(mwidget->document()->findBlockByNumber(i).text().startsWith("//"))) {
            comment = true;
        }
    }
    if (comment) {
        mwidget->prefixLines(sel, "//");
    } else {
        mwidget->chopLines(sel, 2);
    }
}

void TextEditorWidget::keyPressEvent(QKeyEvent *event) {
    if (event->key() == Qt::Key_Tab) {
        Selection sel = currentSelection();
        if (sel.multiline) {
            prefixLines(currentSelection(), "\t");
        } else {
            QPlainTextEdit::keyPressEvent(event);
        }
    } else if (event->key() == Qt::Key_Backtab) {
        Selection sel = currentSelection();
        bool tab = true;
        for (int i = sel.start; i <= sel.end; i++) {
            if (!document()->findBlockByNumber(i).text().startsWith("\t")) {
                tab = false;
            }
        }
        if (tab) {
            chopLines(currentSelection(), 1);
        }
    } else {
        QPlainTextEdit::keyPressEvent(event);
    }
}

Selection TextEditorWidget::currentSelection() {
    Selection ret;
    QTextCursor cursor = textCursor();
    QTextBlock startBlock = document()->findBlock(cursor.selectionStart());
    ret.start = startBlock.blockNumber();
    QTextBlock endBlock = document()->findBlock(cursor.selectionEnd());
    ret.end = endBlock.blockNumber();
    ret.multiline = false;
    if (ret.start != ret.end) {
        ret.multiline = true;
        // don't include line below selection
        if (cursor.selectionEnd() == endBlock.position()) {
            ret.end--;
        }
    }
    return ret;
}

void TextEditorWidget::prefixLines(Selection sel, QString prefix) {
    QTextCursor editCursor(document());
    editCursor.beginEditBlock();
    for (int i = sel.start; i <= sel.end; i++) {
        QTextBlock block = document()->findBlockByNumber(i);
        editCursor.setPosition(block.position());
        editCursor.insertText(prefix);
    }
    editCursor.endEditBlock();
    // select to beginning of first line
    QTextCursor cursor = textCursor();
    QTextBlock startBlock = document()->findBlockByNumber(sel.start);
    int startPos = startBlock.position();
    int endPos = std::max(cursor.position(), cursor.anchor());
    cursor.setPosition(startPos);
    cursor.setPosition(endPos, QTextCursor::KeepAnchor);
    setTextCursor(cursor);
}

void TextEditorWidget::chopLines(Selection sel, int amount) {
    QTextCursor editCursor(document());
    editCursor.beginEditBlock();
    for (int i = sel.start; i <= sel.end; i++) {
        QTextBlock block = document()->findBlockByNumber(i);
        editCursor.setPosition(block.position());
        editCursor.movePosition(QTextCursor::Right, QTextCursor::KeepAnchor, amount);
        editCursor.removeSelectedText();
    }
    editCursor.endEditBlock();
}

BlockInfo TextEditorWidget::topBlock() {
    BlockInfo ret;
    QTextBlock block = firstVisibleBlock();
    ret.index = block.blockNumber();
    ret.top = static_cast<int>(blockBoundingGeometry(block).translated(contentOffset()).top());
    return ret;
}

void TextEditorWidget::resizeEvent(QResizeEvent *e) {
    QPlainTextEdit::resizeEvent(e);
    QRect rect = contentsRect();
    infoStrip->setGeometry(rect.left(), rect.top(), infoStrip->stripWidth(), rect.height());
}

void TextEditorWidget::updateInfoStrip(const QRect &rect, int dy) {
    if (dy > 0) {
        infoStrip->scroll(0, dy);
    } else {
        infoStrip->update(0, rect.y(), infoStrip->width(), rect.height());
    }
}

void TextEditorWidget::contentsModified(bool dirty) {
    errorLines.clear();
    if (dirty) {
        emit modified(this);
    } else {
        emit unmodified(this);
    }
}

void InfoStrip::paintEvent(QPaintEvent *event) {
    QPainter painter(this);
    int lineHeight = fontMetrics().height();
    BlockInfo blockInfo = editor.topBlock();
    int line = blockInfo.index + 1;
    int top = blockInfo.top;
    QSet<int> &errorLines = editor.allErrors();
    while (top < event->rect().bottom() && line <= editor.blockCount()) {
        QRect lineRect(paddingSide, top, width() - paddingSide * 2, lineHeight);
        // markers
        painter.setPen(Qt::red);
        if (errorLines.contains(line)) {
            painter.drawRect(lineRect);
        }
        // line number
        QString number = QString::number(line++);
        painter.setPen(Qt::black);
        painter.drawText(lineRect, Qt::AlignRight, number);
        top += fontMetrics().height();
    }
}

void Editor::warning(QWidget *widget, QString message) {
    QMessageBox::warning(widget, QString("Bipscript IDE"), message);
}
