/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef EDITORWIDGET_H
#define EDITORWIDGET_H

#include "editorlink.h"
#include "namedfile.h"
#include "searchterm.h"

#include <QPlainTextEdit>
#include <QTextDocumentFragment>

class InfoStrip;

struct BlockInfo {
    int index;
    int top;
};

struct Selection {
    int start;
    int end;
    bool multiline;
};

class TextEditorWidget : public QPlainTextEdit {
    Q_OBJECT
    InfoStrip *infoStrip;
    QSet<int> errorLines;

  public:
    TextEditorWidget();
    // error markers
    void addError(int line) { errorLines.insert(line); }
    QSet<int> &allErrors() { return errorLines; }
    BlockInfo topBlock(); // for info strip
    Selection currentSelection();
    void chopLines(Selection sel, int amount);
    void prefixLines(Selection sel, QString prefix);
    void keyPressEvent(QKeyEvent *e) override;

  protected:
    void resizeEvent(QResizeEvent *e) override;

  private slots:
    void updateInfoStrip(const QRect &rect, int dy);
    void contentsModified(bool);

  signals:
    void modified(QWidget *);
    void unmodified(QWidget *);
};

// interface for all editors
class Editor : public QObject {
    NamedFile mfile;

  public:
    Editor(NamedFile &file) : mfile(file) {}
    virtual ~Editor() {}
    const QString &path() { return mfile.path; }
    const QString &name() { return mfile.name; }
    const QString dir() { return mfile.dir(); }
    void moveFile(NamedFile &file) { mfile = file; }
    virtual QWidget *widget() = 0;
    virtual bool isUndoAvailable() { return false; }
    virtual bool isRedoAvailable() { return false; }
    virtual bool isCommentable() { return false; }
    virtual bool isPastable() { return false; }
    virtual bool isRunnable() { return false; }
    virtual bool hasSelection() { return false; }
    virtual void undo() {}
    virtual void redo() {}
    virtual bool isDirty() { return false; }
    virtual void fileChanged() {}
    virtual bool save() = 0;
    virtual bool load() = 0;
    virtual QString selectedText() { return QString(); }

  protected:
    NamedFile &file() { return mfile; }
    void warning(QWidget *widget, QString message);
};

class TextEditor : public Editor {
    Q_OBJECT
    TextEditorWidget *mwidget;
    bool undoable;
    bool redoable;

  public:
    TextEditor(NamedFile &file)
        : Editor(file), mwidget(new TextEditorWidget()), undoable(false), redoable(false) {
        connect(mwidget, SIGNAL(undoAvailable(bool)), this, SLOT(setUndoable(bool)));
        connect(mwidget, SIGNAL(redoAvailable(bool)), this, SLOT(setRedoable(bool)));
    }
    ~TextEditor() { delete mwidget; }
    void focusLine(int line, int column = 0);
    void focusLocation(EditorLocation &loc);
    QList<EditorLink> findAll(const SearchTerm &term, bool active);
    void findNext(const SearchTerm &term);
    void replaceNext(const QString &term, const QString &replace);
    void cut() { mwidget->cut(); }
    void copy() { mwidget->copy(); }
    void paste() { mwidget->paste(); }
    void commentSelection();

    // Editor interface
    QWidget *widget() override { return mwidget; }
    bool isCommentable() override { return true; }
    bool isPastable() override { return true; }
    bool isRunnable() override { return true; }
    bool isDirty() override { return mwidget->document()->isModified(); }
    void fileChanged() override { mwidget->document()->setModified(true); }
    bool save() override;
    bool load() override;
    bool isUndoAvailable() override { return undoable; }
    bool isRedoAvailable() override { return redoable; }
    bool hasSelection() override { return mwidget->textCursor().hasSelection(); }
    QString selectedText() override { return mwidget->textCursor().selection().toPlainText(); }
    void undo() override { mwidget->undo(); }
    void redo() override { mwidget->redo(); }

  public slots:
    void setUndoable(bool b) { undoable = b; }
    void setRedoable(bool b) { redoable = b; }
};

class InfoStrip : public QWidget {
    int paddingSide = 5;
    TextEditorWidget &editor;

  protected:
    void paintEvent(QPaintEvent *event) override;

  public:
    InfoStrip(TextEditorWidget *parent) : QWidget(parent), editor(*parent) {}
    QSize sizeHint() const override { return QSize(stripWidth(), 0); }
    int stripWidth() const {
        return fontMetrics().horizontalAdvance(QLatin1String("9999")) + paddingSide * 2;
    }
};

#endif // EDITORWIDGET_H
