/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "mainwindow.h"

#include "editorwidget.h"
#include "midieditorwidget.h"
#include "projectsdialog.h"
#include "runeditdialog.h"
#include "scriptrun.h"
#include "ui_mainwindow.h"
#include "version.h"

#include <QDebug>
#include <QDesktopServices>
#include <QDir>
#include <QFileDialog>
#include <QInputDialog>
#include <QLabel>
#include <QTime>

MainWindow::MainWindow(QWidget *parent, const char *folder)
    : QMainWindow(parent), ui(new Ui::MainWindow), project(nullptr),
      folderTreeModel(new QFileSystemModel), folderTreeModelProxy(new FolderTreeModelProxy),
      audioEngine(AudioEngine::instance()), transportPosition(0), timeLabel(this), bbtLabel(this) {
    ui->setupUi(this);

    // transport toolbar
    styleGreenMono = "QLabel {font-family: monospace; color: green }";
    styleBlueMono = "QLabel {font-family: monospace; color: blue }";
    styleGreyMono = "QLabel {font-family: monospace; color: gray }";
    // time in MM:SS + BBT + engine config
    timeLabel.setFrameStyle(QFrame::Panel | QFrame::Sunken);
    timeLabel.setStyleSheet(styleGreyMono);
    timeLabel.setText("00:00.000");
    ui->transportToolBar->addWidget(&timeLabel);
    bbtLabel.setFrameStyle(QFrame::Panel | QFrame::Sunken);
    bbtLabel.setStyleSheet(styleGreyMono);
    bbtLabel.setText("--|--|----");
    ui->transportToolBar->addWidget(&bbtLabel);
    engineLabel.setFrameStyle(QFrame::Panel | QFrame::Sunken);
    engineLabel.setStyleSheet(styleGreyMono);
    engineLabel.setText("----- Hz");
    ui->transportToolBar->addWidget(&engineLabel);

    // run toolbar
    runDropdown.setSizeAdjustPolicy(QComboBox::AdjustToContents);
    ui->runToolBar->insertWidget(ui->actionRun, &runDropdown);
    connect(&runDropdown, SIGNAL(currentIndexChanged(int)), this, SLOT(runDropdownChanged(int)));

    // output doc starts hidden
    ui->outputDockWidget->setHidden(true);

    // search widget
    searchWidget = new SearchWidget();
    ui->leftTabWidget->addTab(searchWidget, "Search");
    connect(searchWidget, SIGNAL(searchCurrent(SearchTerm)), this, SLOT(searchCurrent(SearchTerm)));
    connect(searchWidget, SIGNAL(searchNext(SearchTerm)), this, SLOT(searchNext(SearchTerm)));
    connect(searchWidget, SIGNAL(searchProject(SearchTerm)), this, SLOT(searchProject(SearchTerm)));
    connect(searchWidget, SIGNAL(replaceNext(SearchTerm, QString)), this,
            SLOT(replaceNext(SearchTerm, QString)));
    connect(searchWidget, SIGNAL(clearSearch()), this, SLOT(clearSearch()));
    connect(searchWidget, SIGNAL(navigateResult(EditorLocation &)), this,
            SLOT(handleLink(EditorLocation &)));
    connect(&fileWatcher, SIGNAL(fileChanged(QString)), this, SLOT(fileOutsideChange(QString)));

    convertProjectConfig(); // TODO: remove after v0.20

    appConfig.load();
    if (folder) { // from CLI
        Project *project = projectConfig.findByLocation(folder);
        if (project) {
            loadProject(project);
        }
    }
    if (!project) {
        QString pid = appConfig.currentProject();
        if (pid.length()) {
            Project *project = projectConfig.load(pid);
            if (project) {
                loadProject(project);
            }
        }
    }
    if (!project) { // still no project - show projects dialog
        on_actionOpenProject_triggered();
    }
    if (!project) { // still no project! bail out
        QMessageBox::information(this, tr("Goodbye"),
                                 tr("No project selected, application will exit"), QMessageBox::Ok);
        throw "no project, no dice";
    }

    // set up file tree model
    folderTreeModel->setReadOnly(false); // allow file operations via tree control
    folderTreeModelProxy->setSourceModel(folderTreeModel);
    ui->folderTree->setModel(folderTreeModelProxy);
    ui->folderTree->setContextMenuPolicy(Qt::CustomContextMenu);

    connect(ui->folderTree, SIGNAL(customContextMenuRequested(const QPoint &)), this,
            SLOT(fileContextMenu(const QPoint &)));

    connect(ui->folderTree->selectionModel(),
            SIGNAL(currentChanged(const QModelIndex &, const QModelIndex &)), this,
            SLOT(folderTreeSelection(const QModelIndex &, const QModelIndex &)));

    connect(folderTreeModel, SIGNAL(fileRenamed(const QString &, const QString &, const QString &)),
            this, SLOT(fileRenamed(const QString &, const QString &, const QString &)));

    connect(ui->tabWidget, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));

    connect(ui->tabWidget, SIGNAL(tabCloseRequested(int)), this, SLOT(closeEditorTab(int)));
    connect(ui->outputTabWidget, SIGNAL(tabCloseRequested(int)), this, SLOT(closeOutputTab(int)));
}

MainWindow::~MainWindow() { delete ui; }

// ------------ superclass methods

void MainWindow::closeEvent(QCloseEvent *event) {
    saveProject();
    appConfig.save();
    // close tabs
    if (!saveAllTabs()) {
        event->ignore();
    }
    // kill running scripts
    if (!killAllScripts()) {
        event->ignore();
    }
}

// ------------ slots

void MainWindow::changeProject(QString pid) {
    saveProject();
    while (ui->tabWidget->count()) {
        Editor *editor = editorWidgetMap[ui->tabWidget->widget(0)];
        removeEditor(editor->path());
    }
    Project *project = projectConfig.load(pid);
    if (project) {
        loadProject(project);
    }
}

void MainWindow::saveProject() {
    if (project) {
        // save open files
        project->openFiles().clear();
        int filecount = ui->tabWidget->count();
        for (int i = 0; i < filecount; i++) {
            Editor *editor = editorWidgetMap[ui->tabWidget->widget(i)];
            project->openFiles().append(editor->path());
        }
        if (filecount > 0) {
            // save active file
            project->activeFile(currentEditor()->path());
        }
        // save run configs
        project->runConfigs().clear();
        for (int i = 1; i < runDropdown.count(); i++) {
            RunConfig run = runDropdown.itemData(i).value<RunConfig>();
            project->runConfigs().append(run);
        }
        int currentIndex = runDropdown.currentIndex();
        project->activeRun(currentIndex < 0 ? 0 : currentIndex);
        if (!projectConfig.save(*project)) {
            QMessageBox::warning(this, "Project Config", "Failed to save project config");
        }
    }
}

void MainWindow::closeEditorTab(int index) {
    Editor *editor = editorWidgetMap[ui->tabWidget->widget(index)];
    if (saveOrAbandon(editor)) {
        removeEditor(editor->path());
    }
}

void MainWindow::closeOutputTab(int index) {
    OutputWidget *output = qobject_cast<OutputWidget *>(ui->outputTabWidget->widget(index));
    QString name = ui->outputTabWidget->tabText(index);
    if (killOrDetach(output, name)) {
        delete output;
    }
}

void MainWindow::codeError(const QString &path, int line) {
    TextEditorWidget *editor = findTextEditor(path);
    if (editor) {
        editor->addError(line);
    }
}
void MainWindow::engineDisconnect() {
    ui->actionStartStopTransport->setChecked(false);
    enableTransportControls(false);
    audioEngine.disconnect();
    timeLabel.setText("00:00.000");
    bbtLabel.setText("--|--|----");
    ui->actionJackConnect->setChecked(false);
    ui->actionJackConnect->setEnabled(true);
}

void MainWindow::editorModified(QWidget *widget) {
    int index = findTabIndex(widget);
    Editor *editor = editorWidgetMap[widget];
    QString tabText(editor->name());
    tabText.append("*");
    ui->tabWidget->setTabText(index, tabText);
    if (widget == ui->tabWidget->currentWidget()) {
        ui->actionSaveFile->setEnabled(true);
        setWindowModified(true);
    }
}

void MainWindow::editorUnmodified(QWidget *widget) {
    int index = findTabIndex(widget);
    Editor *editor = editorWidgetMap[widget];
    ui->tabWidget->setTabText(index, editor->name());
    if (widget == ui->tabWidget->currentWidget()) {
        ui->actionSaveFile->setEnabled(false);
        setWindowModified(false);
    }
}

void MainWindow::fileContextMenu(const QPoint &point) {
    QModelIndex index = ui->folderTree->indexAt(point);
    if (index.isValid()) {
        QModelIndex current = folderTreeModelProxy->mapToSource(index);
        QString path = folderTreeModel->filePath(current);
        QFileInfo fileInfo(path);

        if (fileInfo.isFile()) {
            QMenu contextMenu(tr("File menu"), this);

            // info
            QAction item1("Info", &contextMenu);
            connect(&item1, SIGNAL(triggered()), this, SLOT(fileInfo()));
            contextMenu.addAction(&item1);

            // run
            QAction item2("Run", &contextMenu);
            connect(&item2, SIGNAL(triggered()), this, SLOT(on_actionRun_triggered()));
            contextMenu.addAction(&item2);

            // run w/ args
            QAction item3("Run with Arguments", &contextMenu);
            connect(&item3, SIGNAL(triggered()), this, SLOT(runScriptWithArgs()));
            contextMenu.addAction(&item3);

            // delete
            //            QAction item4("Delete", &contextMenu);
            //            item4.setIcon(QIcon::fromTheme("edit-delete"));
            //            connect(&item4, SIGNAL(triggered()), this,
            //            SLOT(on_actionDeleteFile_triggered())); contextMenu.addAction(&item4);

            contextMenu.exec(ui->folderTree->mapToGlobal(point));

        } else if (fileInfo.isDir()) {
            QMenu contextMenu(tr("File menu"), this);

            // new file
            QAction item1("New File", &contextMenu);
            connect(&item1, SIGNAL(triggered()), this, SLOT(on_actionNewFile_triggered()));
            contextMenu.addAction(&item1);

            // new folder
            //            QAction item2("New Folder",  &contextMenu);
            //            connect(&item1, SIGNAL(triggered()), this,
            //            SLOT(on_actionNewFolder_triggered())); contextMenu.addAction(&item2);

            contextMenu.exec(ui->folderTree->mapToGlobal(point));
        }
    }
}

void MainWindow::fileInfo() {
    QFileInfo fileinfo(currentEditor()->path());
    QString created = fileinfo.birthTime().toString();
    QString modified = fileinfo.lastModified().toString();
    QMessageBox::information(this, tr("File Info"),
                             tr("filename: <b>%1</b><br/>created: %2<br/>modified: %3")
                                 .arg(fileinfo.fileName(), created, modified),
                             QMessageBox::Ok);
}

// TODO: only handles text editors, should handle others
void MainWindow::fileOutsideChange(QString path) {
    Editor *editor = editorPathMap[path];
    if (editor) {
        QMessageBox::StandardButton choice = QMessageBox::warning(
            this, tr("File Changed"),
            tr("File %1 has changed outside of the IDE.\nDo you wish to reload from disk?")
                .arg(path),
            QMessageBox::Yes | QMessageBox::Cancel);
        if (choice == QMessageBox::Yes) {
            // block signals to prevent double-firing
            // TODO: is this the best way to handle this?
            fileWatcher.blockSignals(true);
            editor->load();
            fileWatcher.blockSignals(false);
        } else {
            editor->fileChanged();
        }
    }
}

void MainWindow::fileRenamed(const QString &folder, const QString &oldname,
                             const QString &newname) {

    // update project
    NamedFile oldfile(oldname, folder + '/' + oldname);
    fileWatcher.removePath(oldfile.path);
    NamedFile file(newname, folder + '/' + newname);
    // update editor
    currentEditor()->moveFile(file);
    fileWatcher.addPath(file.path);
    editorPathMap[file.path] = currentEditor();
    editorPathMap.remove(oldfile.path);
    // update tab
    QString tabText(currentEditor()->name());
    if (currentEditor()->isDirty()) {
        tabText.append("*");
    }
    int tabindex = ui->tabWidget->currentIndex();
    ui->tabWidget->setTabText(tabindex, tabText);
    // update main window title
    setWindowTitle(QString("%1[*]").arg(file.name));
}

void MainWindow::addRunToList(RunConfig &run) {
    QList<RunConfig> &configs = project->runConfigs();
    if (!configs.contains(run)) {
        configs.append(run);
        QVariant data;
        data.setValue(run);
        runDropdown.addItem(run.name(), data);
    }
}

bool MainWindow::createOrOpenEditor(NamedFile &file) {
    Editor *editor = editorPathMap[file.path];
    if (editor) {
        ui->tabWidget->setCurrentWidget(editor->widget());
    } else { // new editor for this file
        QWidget *ewidget = nullptr;
        if (file.path.endsWith(".mid") || file.path.endsWith(".midi")) {
            editor = new MidiEditor(file);
            ewidget = editor->widget();
            if (!editor->load()) {
                return false;
            }
        } else {
            editor = new TextEditor(file);
            ewidget = editor->widget();
            if (editor->load()) {
                connect(ewidget, SIGNAL(undoAvailable(bool)), ui->actionUndo,
                        SLOT(setEnabled(bool)));
                connect(ewidget, SIGNAL(redoAvailable(bool)), ui->actionRedo,
                        SLOT(setEnabled(bool)));
                connect(ewidget, SIGNAL(copyAvailable(bool)), ui->actionCut,
                        SLOT(setEnabled(bool)));
                connect(ewidget, SIGNAL(copyAvailable(bool)), ui->actionCopy,
                        SLOT(setEnabled(bool)));
                connect(ewidget, SIGNAL(modified(QWidget *)), this,
                        SLOT(editorModified(QWidget *)));
                connect(ewidget, SIGNAL(unmodified(QWidget *)), this,
                        SLOT(editorUnmodified(QWidget *)));
                // add to project list as open file
            } else {
                return false;
            }
        }
        fileWatcher.addPath(file.path);
        ui->tabWidget->addTab(ewidget, file.name);
        ui->tabWidget->setCurrentWidget(ewidget);
        // add to maps
        editorPathMap[file.path] = editor;
        editorWidgetMap[editor->widget()] = editor;
    }
    enableFileActions(editor);
    return true;
}

bool MainWindow::createOrOpenEditor(const QString &path) {
    QFileInfo fileInfo(path);
    NamedFile file(fileInfo.fileName(), fileInfo.filePath());
    return createOrOpenEditor(file);
}

Editor *MainWindow::currentEditor() { return editorWidgetMap[ui->tabWidget->currentWidget()]; }

TextEditor *MainWindow::currentTextEditor() { return dynamic_cast<TextEditor *>(currentEditor()); }

void MainWindow::folderTreeSelection(const QModelIndex &pcurrent, const QModelIndex &) {
    if (pcurrent.isValid() && pcurrent.model() == folderTreeModelProxy) {
        QModelIndex current = folderTreeModelProxy->mapToSource(pcurrent);
        QString path = folderTreeModel->filePath(current);
        QFileInfo fileInfo(path);
        if (fileInfo.isDir()) {
            enableFileActions(nullptr);
            ui->actionNewFile->setEnabled(true);
        } else {
            ui->actionNewFile->setEnabled(false);
            if (fileInfo.isFile()) {
                createOrOpenEditor(path);
            } else {
                QMessageBox::information(this, "Unknown file type", path);
            }
        }
    }
}

void MainWindow::handleLink(const QUrl &url) {
    QString path(url.path());
    createOrOpenEditor(path);
    if (url.hasQuery()) {
        int column = QString(url.query()).toInt();
        currentTextEditor()->focusLine(url.port(), column);
    } else {
        currentTextEditor()->focusLine(url.port());
    }
}

void MainWindow::handleLink(EditorLocation &loc) {
    createOrOpenEditor(loc.file().path);
    currentTextEditor()->focusLocation(loc);
}

void MainWindow::runDropdownChanged(int index) {
    ui->actionDeleteRun->setEnabled(index);
    ui->actionEditRun->setEnabled(index);
}

void MainWindow::searchCurrent(const SearchTerm &search) {
    TextEditor *editor = currentTextEditor();
    if (editor) {
        searchWidget->addResults(editor->path(), editor->name(), editor->findAll(search, true),
                                 true);
    }
}

void MainWindow::searchNext(const SearchTerm &search) {
    if (currentTextEditor()) {
        currentTextEditor()->findNext(search);
    }
}

void MainWindow::replaceNext(const SearchTerm &search, const QString &replace) {
    TextEditor *editor = currentTextEditor();
    if (editor) {
        editor->replaceNext(search.string(), replace);
        searchWidget->updateResults(editor->path(), editor->findAll(search, true));
    }
}

void MainWindow::runScriptWithArgs() {
    QString message = tr("Command line arguments for %1").arg(currentEditor()->name());
    bool ok;
    QString args =
        QInputDialog::getText(this, "Runtime Arguments", message, QLineEdit::Normal, "", &ok);
    if (ok) {
        QString path = currentEditor()->path();
        runScript(path, args);
        RunConfig run(path, args);
        addRunToList(run);
    }
}

void MainWindow::searchFolder(const SearchTerm &search, QString path) {
    QDir dir(path);
    dir.setFilter(QDir::AllEntries | QDir::NoDotAndDotDot);

    foreach (QFileInfo fileInfo, dir.entryInfoList()) {
        if (fileInfo.isFile()) {
            TextEditor *editor = dynamic_cast<TextEditor *>(editorPathMap[fileInfo.filePath()]);
            if (editor) {
                QList<EditorLink> results = editor->findAll(search, false);
                if (results.size()) {
                    searchWidget->addResults(editor->path(), editor->name(), results, true);
                }
            } else {
                QList<EditorLink> results = searchFile(search, fileInfo.filePath());
                if (results.size()) {
                    searchWidget->addResults(fileInfo.filePath(), fileInfo.fileName(), results,
                                             false);
                }
            }
        } else if (fileInfo.isDir()) {
            searchFolder(search, fileInfo.filePath());
        }
    }
}

void MainWindow::searchProject(const SearchTerm &search) {
    for (QString folder : project->folders()) {
        searchFolder(search, folder);
    }
}

void MainWindow::clearSearch() {
    for (int i = 0; i < ui->tabWidget->count(); i++) {
        TextEditorWidget *editor = qobject_cast<TextEditorWidget *>(ui->tabWidget->widget(i));
        QList<QTextEdit::ExtraSelection> empty;
        editor->setExtraSelections(empty);
    }
}

void MainWindow::showTime(Position pos) {
    QTime qtime(0, 0, 0);
    qtime = qtime.addMSecs(pos.msec);
    timeLabel.setText(qtime.toString("mm:ss.zzz"));
    bbtLabel.setText(QString("%1|%2|%3").arg(pos.bar, 2).arg(pos.beat, 2).arg(pos.tick, 4));
    transportPosition = pos.frame;
}

void MainWindow::tabChanged(int i) {
    Editor *editor = editorWidgetMap[ui->tabWidget->widget(i)];
    if (editor == nullptr) { // all tabs closed
        ui->folderTree->selectionModel()->clear();
        setWindowTitle(QCoreApplication::applicationName() + " [*]");
        enableFileActions(nullptr);
    } else {
        QModelIndex index = folderTreeModelIndex(editor->path());
        ui->folderTree->setCurrentIndex(index);
        setWindowTitle(QString("%1[*]").arg(editor->name()));
        setWindowModified(editor->isDirty());
        enableFileActions(editor);
    }
}

void MainWindow::closeOtherEditors() {
    if (QAction *action = qobject_cast<QAction *>(sender())) {
        int index = action->data().toInt();
        for (int i = ui->tabWidget->count() - 1; i >= 0; i--) {
            if (i != index) {
                closeEditorTab(i);
            }
        }
    }
}

void MainWindow::closeOtherOutputs() {
    if (QAction *action = qobject_cast<QAction *>(sender())) {
        int index = action->data().toInt();
        for (int i = ui->outputTabWidget->count() - 1; i >= 0; i--) {
            if (i != index) {
                closeOutputTab(i);
            }
        }
    }
}

// ------------ auto slots

void MainWindow::on_actionAbout_triggered() {
    QString lang = QLocale::languageToString(QLocale().language());
    QString message =
        tr("Bipscript IDE %1\n%2\nQT %3").arg(BIPSCRIPT_IDE_VERSION).arg(lang).arg(QT_VERSION_STR);
    QMessageBox::about(this, "Bipscript-IDE", message);
}

void MainWindow::on_actionAPIDocumentation_triggered() {
    QStringList supported = {"en", "es"};
    QString lang = QLocale().name().left(2);
    if (supported.contains(lang)) {
        lang = supported[0];
    }
    QString link = QString("http://www.bipscript.org/%1/api/").arg(lang);
    QDesktopServices::openUrl(QUrl(link));
}

void MainWindow::on_actionCloseFile_triggered() { closeEditorTab(ui->tabWidget->currentIndex()); }

void MainWindow::on_actionComment_triggered() { currentTextEditor()->commentSelection(); }

void MainWindow::on_actionCopy_triggered() { currentTextEditor()->copy(); }

void MainWindow::on_actionCut_triggered() { currentTextEditor()->cut(); }

void MainWindow::on_actionNewFile_triggered() {
    QModelIndex index = ui->folderTree->selectionModel()->currentIndex();
    QString path = folderTreeModel->filePath(folderTreeModelProxy->mapToSource(index));
    QString message("New file in ");
    message.append(path);
    bool ok;
    QString name = QInputDialog::getText(this, "New File", message, QLineEdit::Normal, "", &ok);
    if (name.length() > 0) {
        // create file
        QFileInfo fileInfo(path + "/" + name);
        QString path(fileInfo.filePath());
        QFile file(path);
        if (file.exists()) {
            QMessageBox::warning(this, "File Exists", path + " already exists");
        } else {
            file.open(QIODevice::WriteOnly);
            file.close();
        }
        // select in tree
        QModelIndex index = folderTreeModelIndex(path);
        ui->folderTree->setCurrentIndex(index);
        ui->folderTree->selectionModel()->select(index, QItemSelectionModel::Select);
    }
}

void MainWindow::on_actionDeleteFile_triggered() {
    QMessageBox msgBox;
    QString path(currentEditor()->path());
    msgBox.setText("Delete " + path);
    msgBox.setInformativeText("Do you want to delete " + currentEditor()->name() + "?");
    msgBox.setStandardButtons(QMessageBox::Ok | QMessageBox::Cancel);
    msgBox.setDefaultButton(QMessageBox::Cancel);
    int choice = msgBox.exec();
    if (choice == QMessageBox::Ok) {
        // remove editor if any
        removeEditor(path);
        // remove file from filesystem
        QModelIndex index = folderTreeModel->index(path);
        folderTreeModel->remove(index);
    }
}

void MainWindow::on_actionDeleteRun_triggered() {
    runDropdown.removeItem(runDropdown.currentIndex());
}

void MainWindow::on_actionEditRun_triggered() {
    RunConfig run = runDropdown.currentData().value<RunConfig>();
    RunEditDialog dialog(this, run);
    int result = dialog.exec();
    if (result == QDialog::Accepted) {
        // OK = save
        QVariant data;
        data.setValue(run);
        int idx = runDropdown.currentIndex();
        runDropdown.setItemData(idx, data);
        runDropdown.setItemText(idx, run.name());
    }
}

void MainWindow::on_actionExamples_triggered() {
    QRegExp versionExp = QRegExp("v([0-9\\.]+)");
    QString versionString = QString(BIPSCRIPT_IDE_VERSION);
    if (versionExp.indexIn(versionString) == 0) {
        QString version = versionExp.cap(1);
        QString link("https://gitlab.domainepublic.net/bipscript/examples/-/releases/v");
        link.append(version);
        QDesktopServices::openUrl(QUrl(link));
    } else {
        QString message = tr("No examples for version %1").arg(versionString);
        QMessageBox::warning(this, tr("Unknown Version"), message, QMessageBox::Ok);
    }
}

void MainWindow::on_actionExit_triggered() { QApplication::closeAllWindows(); }

void MainWindow::on_actionFastForward_triggered() {
    transportPosition = transportPosition + audioConfig.framesPerSecond;
    audioEngine.reposition(transportPosition);
}

void MainWindow::on_actionFind_in_Project_triggered() {
    searchWidget->newProjectSearch(currentTextEditor()->selectedText());
    ui->leftTabWidget->setCurrentWidget(searchWidget);
}

/**
 * @brief MainWindow::on_actionRun_triggered
 *
 * action to run script
 */
void MainWindow::on_actionRun_triggered() {
    if (runDropdown.currentIndex() == 0) {
        // run current active file with no args
        runScript(currentEditor()->path(), "");
        // create new run config and save to list
        RunConfig run(currentEditor()->path(), "");
        addRunToList(run);
    } else {
        RunConfig run = runDropdown.currentData().value<RunConfig>();
        runScript(run.path().canonicalFilePath(), run.args());
    }
}

void MainWindow::enableTransportControls(bool enabled) {
    ui->actionStartStopTransport->setEnabled(enabled);
    ui->actionRewind->setEnabled(enabled);
    ui->actionRestart->setEnabled(enabled);
    ui->actionFastForward->setEnabled(enabled);
    if (enabled) {
        timeLabel.setStyleSheet(styleGreenMono);
        bbtLabel.setStyleSheet(styleGreenMono);
        engineLabel.setStyleSheet(styleBlueMono);
    } else {
        timeLabel.setStyleSheet(styleGreyMono);
        bbtLabel.setStyleSheet(styleGreyMono);
        engineLabel.setStyleSheet(styleGreyMono);
    }
}

void MainWindow::on_actionJackConnect_toggled(bool connected) {
    if (connected) {
        audioConfig = audioEngine.connect();
        atimer = audioEngine.timerService();
        connect(atimer, SIGNAL(transport(Position)), this, SLOT(showTime(Position)));
        connect(atimer, SIGNAL(rolling(bool)), ui->actionStartStopTransport,
                SLOT(setChecked(bool)));
        connect(atimer, SIGNAL(finished()), this, SLOT(engineDisconnect()));
        engineLabel.setText(QString("%1 Hz").arg(audioConfig.framesPerSecond));
    } else {
        ui->actionJackConnect->setEnabled(false);
        atimer->stop();
        engineLabel.setText("----- Hz");
    }
    enableTransportControls(connected && !ui->actionTransportLock->isChecked());
}

void MainWindow::on_actionOpenFile_triggered() {
    QFileDialog fileDialog(this);
    fileDialog.setFileMode(QFileDialog::AnyFile);
    Editor *current = currentEditor();
    if (current) {
        fileDialog.setDirectory(currentEditor()->dir());
    } else {
        fileDialog.setDirectory(QDir::homePath());
    }
    if (fileDialog.exec()) {
        QString path = fileDialog.selectedFiles()[0];
        createOrOpenEditor(path);
    }
}

void MainWindow::on_actionOpenProject_triggered() {
    if (saveAllTabs()) {
        ProjectsDialog projectsDialog(this, project);
        connect(&projectsDialog, SIGNAL(projectChanged(QString)), this,
                SLOT(changeProject(QString)));
        projectsDialog.exec();
    }
}

void MainWindow::on_actionPaste_triggered() { currentTextEditor()->paste(); }

void MainWindow::on_actionProjectOptions_triggered() {
    QString value = project->executable();
    QString message("Executable Path");
    bool ok;
    QString name = QInputDialog::getText(this, "Options", message, QLineEdit::Normal, value, &ok);
    if (name.length() > 0) {
        // TODO: test existance etc.
        project->executable(name);
        ScriptRun::setExecutable(project->executable());
    }
}

void MainWindow::on_actionRedo_triggered() { currentEditor()->redo(); }

void MainWindow::on_actionRestart_triggered() { audioEngine.reposition(0); }

void MainWindow::on_actionRewind_triggered() {
    int newpos = static_cast<int>(transportPosition - audioConfig.framesPerSecond);
    transportPosition = static_cast<uint>(qMax(newpos, 0));
    audioEngine.reposition(transportPosition);
}

void MainWindow::on_actionSaveFile_triggered() { saveEditor(currentEditor()); }

void MainWindow::on_actionFocusFind_triggered() {
    searchWidget->newFileSearch(currentEditor()->selectedText());
    ui->leftTabWidget->setCurrentWidget(searchWidget);
}

void MainWindow::on_actionStartStopTransport_toggled(bool start) {
    if (start) {
        audioEngine.transportStart();
    } else {
        audioEngine.transportStop();
    }
    ui->actionJackConnect->setEnabled(!start);
}

void MainWindow::on_actionTransportLock_toggled(bool locked) {
    bool enabled = !locked && ui->actionJackConnect->isChecked();
    ui->actionStartStopTransport->setEnabled(enabled);
    ui->actionRewind->setEnabled(enabled);
    ui->actionRestart->setEnabled(enabled);
    ui->actionFastForward->setEnabled(enabled);
}

void MainWindow::on_actionUndo_triggered() { currentEditor()->undo(); }

void MainWindow::on_tabWidget_customContextMenuRequested(const QPoint &pos) {
    QMenu menu;
    QAction *action = menu.addAction(tr("Close Other Tabs"), this, SLOT(closeOtherEditors()));
    action->setData(ui->tabWidget->tabBar()->tabAt(pos));
    action->setEnabled(ui->tabWidget->count() > 1);
    menu.exec(QCursor::pos());
}

// ------------ private methods

void MainWindow::enableFileActions(Editor *editor) {
    // requires file selected
    ui->actionDeleteFile->setEnabled(editor);
    ui->actionCloseFile->setEnabled(editor);
    // requires state from the editor
    ui->actionRun->setEnabled(runDropdown.currentIndex() || (editor && editor->isRunnable()));
    ui->actionComment->setEnabled(editor && editor->isCommentable());
    ui->actionSaveFile->setEnabled(editor && editor->isDirty());
    ui->actionPaste->setEnabled(editor && editor->isPastable());
    ui->actionUndo->setEnabled(editor && editor->isUndoAvailable());
    ui->actionRedo->setEnabled(editor && editor->isRedoAvailable());
    ui->actionCut->setEnabled(editor && editor->hasSelection());
    ui->actionCopy->setEnabled(editor && editor->hasSelection());
}

TextEditorWidget *MainWindow::findTextEditor(const QString &path) {
    TextEditor *editor = dynamic_cast<TextEditor *>(editorPathMap[path]);
    return editor ? static_cast<TextEditorWidget *>(editor->widget()) : nullptr;
}

int MainWindow::findTabIndex(QWidget *widget) {
    for (int i = 0; i < ui->tabWidget->count(); i++) {
        if (ui->tabWidget->widget(i) == widget) {
            return i;
        }
    }
    return -1;
}

QModelIndex MainWindow::folderTreeModelIndex(QString path) {
    QModelIndex idx = folderTreeModel->index(path);
    return folderTreeModelProxy->mapFromSource(idx);
}

void MainWindow::convertProjectConfig() {
    ProjectList &projectList = ProjectList::instance();
    projectList.load();
    if (projectConfig.count() == 0) {
        for (int i = 0; i < projectList.size(); i++) {
            LegacyProject &legacy = projectList.get(i);
            // old to new
            Project *project = new Project(projectConfig.createId());
            project->name(legacy.folder(0).name); // make a name
            project->executable(legacy.executable());
            project->activeFile(legacy.activeFile());
            for (int i = 0; i < legacy.folderCount(); i++) {
                project->folders().append(legacy.folder(i).path);
            }
            for (QString open : legacy.openFiles()) {
                project->folders().append(open);
            }
            projectConfig.save(*project);
        }
    }
}

void MainWindow::loadProject(Project *inproject) {
    // set project
    this->project = inproject;
    // set intrepreter executable
    ScriptRun::setExecutable(project->executable());

    // populate file tree
    QString path = project->folders()[0];
    QFileInfo fileinfo(path);
    QString parent = fileinfo.dir().path();
    folderTreeModel->setRootPath(parent);
    folderTreeModelProxy = new FolderTreeModelProxy(path);
    folderTreeModelProxy->setSourceModel(folderTreeModel);
    ui->folderTree->setModel(folderTreeModelProxy);
    ui->folderTree->setRootIndex(folderTreeModelIndex(parent));

    connect(ui->folderTree->selectionModel(),
            SIGNAL(currentChanged(const QModelIndex &, const QModelIndex &)), this,
            SLOT(folderTreeSelection(const QModelIndex &, const QModelIndex &)));

    // hide extra columns - file date, etc
    ui->folderTree->header()->hideSection(1);
    ui->folderTree->header()->hideSection(2);
    ui->folderTree->header()->hideSection(3);

    // currently open files for this project
    foreach (const QString &path, project->openFiles()) {
        createOrOpenEditor(path);
    }

    // run dropdown
    runDropdown.clear();
    runDropdown.addItem("[current]");
    QList<RunConfig> &runs = project->runConfigs();
    for (int i = 0; i < runs.size(); i++) {
        QVariant data;
        data.setValue(runs.at(i));
        runDropdown.addItem(runs.at(i).name(), data);
    }
    runDropdown.setCurrentIndex(project->activeRun());

    // set currently showing editor
    Editor *showing = editorPathMap[project->activeFile()];
    if (showing) {
        ui->tabWidget->setCurrentWidget(showing->widget());
        setWindowTitle(QString("%1[*]").arg(showing->name()));
        QModelIndex index = folderTreeModelIndex(showing->path());
        ui->folderTree->setCurrentIndex(index);
        enableFileActions(showing);
    }
    // reset search widget and show tree
    searchWidget->clear();
    ui->leftTabWidget->setCurrentWidget(ui->folderTree);
    // store as current project
    appConfig.currentProject(project->id());
}

void MainWindow::removeEditor(QString path) {
    if (editorPathMap.find(path) != editorPathMap.end()) {
        Editor *editor = editorPathMap[path];
        editorPathMap.remove(path);
        editorWidgetMap.remove(editor->widget());
        fileWatcher.removePath(editor->path());
        delete editor;
    }
}

void MainWindow::runScript(QString path, QString args) {
    // check if possible
    if (!ScriptRun::possible()) {
        QMessageBox::warning(this, "Executable", "Cannot launch bipscript executable");
        return;
    }
    // add output widget for this run
    OutputWidget *out = new OutputWidget();
    ScriptRun *run = new ScriptRun(path, args, out->port());
    QFileInfo fileinfo(path);
    ui->outputTabWidget->addTab(out, fileinfo.fileName());
    ui->outputTabWidget->setCurrentWidget(out);
    connect(run, SIGNAL(started(int)), out, SLOT(running(int)));
    connect(run, SIGNAL(readStdout(const char *)), out, SLOT(write(const char *)));
    connect(run, SIGNAL(readStderr(QString)), out, SLOT(writeError(QString)));
    connect(run, SIGNAL(error(QString)), out, SLOT(writeError(QString)));
    connect(run, SIGNAL(exited(int)), out, SLOT(completed(int)));
    connect(out, SIGNAL(textInput(QString)), run, SLOT(writeStdin(QString)));
    connect(out, SIGNAL(codeError(const QString &, int)), this,
            SLOT(codeError(const QString &, int)));
    connect(out, SIGNAL(linkClicked(const QUrl &)), this, SLOT(handleLink(const QUrl &)));
    // show output area if hidden
    ui->outputDockWidget->show();
    run->start();
}

bool MainWindow::saveEditor(Editor *editor) {
    // suspend file watcher for save
    fileWatcher.removePath(editor->path());
    bool result = editor->save();
    fileWatcher.addPath(editor->path());
    return result;
}

bool MainWindow::saveAllTabs() {
    for (int i = 0; i < ui->tabWidget->count(); i++) {
        Editor *editor = editorWidgetMap[ui->tabWidget->widget(i)];
        if (editor->isDirty()) {
            QMessageBox::StandardButton choice = unsavedChangesDialog(editor->path());
            if (choice == QMessageBox::Cancel) {
                return false;
            } else if (choice == QMessageBox::Save) {
                saveEditor(editor);
            }
        }
    }
    return true;
}

bool MainWindow::saveOrAbandon(Editor *editor) {
    if (editor->isDirty()) {
        QMessageBox::StandardButton choice = unsavedChangesDialog(editor->path());
        if (choice == QMessageBox::Cancel) {
            return false;
        } else if (choice == QMessageBox::Save) {
            if (!saveEditor(editor)) {
                return false;
            }
        }
    }
    return true;
}

QList<EditorLink> MainWindow::searchFile(const SearchTerm &search, const QString &path) {
    QList<EditorLink> ret;
    QFile file(path);
    Qt::CaseSensitivity sensitivity =
        search.caseSensitive() ? Qt::CaseSensitive : Qt::CaseInsensitive;
    QString term = search.string();
    if (file.open(QFile::ReadOnly | QFile::Text)) {
        QTextStream in(&file);
        int lineindex = 0;
        while (!in.atEnd()) {
            QString line = in.readLine();
            int index = line.indexOf(term, 0, sensitivity);
            while (index != -1) {
                NamedFile nfile(file.fileName(), path);
                int anchor = lineindex + index;
                // find snippet
                int space = line.indexOf(QRegExp("\\s+"), index + term.size() + 1);
                int end = (space == -1 ? line.size() : space);
                QString text = line.mid(index, end - index);
                // create and add editor link
                EditorLink link(text, nfile, anchor, anchor + term.size());
                ret.append(link);
                index = line.indexOf(term, index + 1, sensitivity);
            }
            lineindex += line.length() + 1;
        }
    }
    return ret;
}

bool MainWindow::killAllScripts() {
    for (int i = 0; i < ui->outputTabWidget->count(); i++) {
        OutputWidget *output = qobject_cast<OutputWidget *>(ui->outputTabWidget->widget(i));
        QString name = ui->outputTabWidget->tabText(i);
        if (!killOrDetach(output, name)) {
            return false;
        }
    }
    return true;
}

bool MainWindow::killOrDetach(OutputWidget *output, const QString &name) {
    if (output->pid() != 0) {
        QString id = QString("%1 pid %2").arg(name).arg(output->pid());
        QString mesg = "%1 is still running.\nDo you wish to kill the process?";
        QMessageBox::StandardButton choice =
            QMessageBox::warning(this, tr("Running Process"), mesg.arg(id),
                                 QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
        if (choice == QMessageBox::Cancel) {
            return false;
        } else if (choice == QMessageBox::Yes) {
            output->endScript(true);
        }
    }
    return true;
}

QMessageBox::StandardButton MainWindow::unsavedChangesDialog(const QString &path) {
    return QMessageBox::warning(this, tr("Unsaved Changes"),
                                tr("%1 has unsaved changes.\nDo you wish to save?").arg(path),
                                QMessageBox::Save | QMessageBox::No | QMessageBox::Cancel);
}

void MainWindow::on_actionWebsite_triggered() {
    QString link = "http://www.bipscript.org";
    QDesktopServices::openUrl(QUrl(link));
}

void MainWindow::on_outputTabWidget_customContextMenuRequested(const QPoint &pos) {
    QMenu menu;
    QAction *action = menu.addAction(tr("Close Other Tabs"), this, SLOT(closeOtherOutputs()));
    action->setData(ui->outputTabWidget->tabBar()->tabAt(pos));
    action->setEnabled(ui->outputTabWidget->count() > 1);
    menu.exec(QCursor::pos());
}

bool FolderTreeModelProxy::filterAcceptsRow(int row, const QModelIndex &parent) const {
    QFileSystemModel *fileModel = qobject_cast<QFileSystemModel *>(sourceModel());
    QModelIndex index = fileModel->index(row, 0, parent);
    if (fileModel) {
        QString path = fileModel->filePath(index);
        QFileInfo fileinfo(fileModel->filePath(index));
        // only show the project root
        if (fileinfo.absoluteDir() == mroot.absoluteDir()) {
            return fileinfo.canonicalFilePath() == mroot.canonicalFilePath();
        }
    }
    return true;
}
