/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "applicationconfig.h"
#include "audioengine.h"
#include "editorwidget.h"
#include "outputwidget.h"
#include "position.h"
#include "projectconfig.h"
#include "projectlist.h"
#include "searchwidget.h"

#include <QComboBox>
#include <QFileInfo>
#include <QFileSystemModel>
#include <QFileSystemWatcher>
#include <QLabel>
#include <QMainWindow>
#include <QMessageBox>
#include <QSortFilterProxyModel>

namespace Ui {
class MainWindow;
}

class FolderTreeModelProxy : public QSortFilterProxyModel {
    QFileInfo mroot;

  public:
    FolderTreeModelProxy() {}
    FolderTreeModelProxy(QString root) : mroot(root) {}
    bool filterAcceptsRow(int sourceRow, const QModelIndex &sourceParent) const override;
};

class MainWindow : public QMainWindow {
    Q_OBJECT
  public:
    explicit MainWindow(QWidget *parent = 0, const char *project = nullptr);
    ~MainWindow();
    void closeEvent(QCloseEvent *event);

  private slots:
    void changeProject(QString id);
    void clearSearch();
    void closeOutputTab(int index);
    void closeOtherOutputs();
    void closeEditorTab(int index);
    void closeOtherEditors();
    void codeError(const QString &, int);
    void engineDisconnect();
    void editorModified(QWidget *widget);
    void editorUnmodified(QWidget *widget);
    void fileContextMenu(const QPoint &);
    void fileInfo();
    void fileOutsideChange(QString);
    void fileRenamed(const QString &, const QString &, const QString &);
    void folderTreeSelection(const QModelIndex &, const QModelIndex &);
    void handleLink(const QUrl &);
    void handleLink(EditorLocation &);
    void runDropdownChanged(int index);
    void searchCurrent(const SearchTerm &);
    void searchNext(const SearchTerm &);
    void searchProject(const SearchTerm &);
    void replaceNext(const SearchTerm &term, const QString &replace);
    void runScriptWithArgs();
    void showTime(Position pos);
    void tabChanged(int index);
    void on_actionAbout_triggered();
    void on_actionAPIDocumentation_triggered();
    void on_actionCloseFile_triggered();
    void on_actionComment_triggered();
    void on_actionCopy_triggered();
    void on_actionCut_triggered();
    void on_actionDeleteFile_triggered();
    void on_actionDeleteRun_triggered();
    void on_actionEditRun_triggered();
    void on_actionExamples_triggered();
    void on_actionExit_triggered();
    void on_actionFastForward_triggered();
    void on_actionFind_in_Project_triggered();
    void on_actionFocusFind_triggered();
    void on_actionJackConnect_toggled(bool);
    void on_actionNewFile_triggered();
    void on_actionOpenFile_triggered();
    void on_actionOpenProject_triggered();
    void on_actionPaste_triggered();
    void on_actionProjectOptions_triggered();
    void on_actionRedo_triggered();
    void on_actionRestart_triggered();
    void on_actionRewind_triggered();
    void on_actionRun_triggered();
    void on_actionSaveFile_triggered();
    void on_actionStartStopTransport_toggled(bool);
    void on_actionTransportLock_toggled(bool);
    void on_actionUndo_triggered();
    void on_actionWebsite_triggered();
    void on_tabWidget_customContextMenuRequested(const QPoint &pos);
    void on_outputTabWidget_customContextMenuRequested(const QPoint &pos);

  private:
    Ui::MainWindow *ui;
    ApplicationConfig appConfig;
    ProjectConfig projectConfig;
    Project *project;
    QFileSystemModel *folderTreeModel;
    FolderTreeModelProxy *folderTreeModelProxy;

    // editors
    QMap<QString, Editor *> editorPathMap;
    QMap<QWidget *, Editor *> editorWidgetMap;

    // audio transport stuff - separate component?
    AudioEngine &audioEngine;
    AudioTimerService *atimer;
    uint transportPosition;
    AudioConfiguration audioConfig;
    QLabel timeLabel;
    QLabel bbtLabel;
    QLabel engineLabel;
    QString styleGreenMono;
    QString styleBlueMono;
    QString styleGreyMono;

    // run toolbar
    QComboBox runDropdown;

    SearchWidget *searchWidget;     // file text search
    QFileSystemWatcher fileWatcher; // watch files for changes

    // private methods
    void addRunToList(RunConfig &run);
    bool createOrOpenEditor(NamedFile &);
    bool createOrOpenEditor(const QString &);
    Editor *currentEditor();
    TextEditor *currentTextEditor();
    void enableFileActions(Editor *editor);
    void enableTransportControls(bool enabled);
    int findTabIndex(QWidget *widget);
    TextEditorWidget *findTextEditor(const QString &path);
    QModelIndex folderTreeModelIndex(QString path);
    bool killAllScripts();
    bool killOrDetach(OutputWidget *output, const QString &name);
    void loadProject(Project *project);
    void removeEditor(QString path);
    void runScript(QString path, QString args);
    bool saveEditor(Editor *);
    bool saveAllTabs();
    bool saveOrAbandon(Editor *editor);
    QList<EditorLink> searchFile(const SearchTerm &term, const QString &path);
    void searchFolder(const SearchTerm &term, QString path);
    QMessageBox::StandardButton unsavedChangesDialog(const QString &);
    void convertProjectConfig(); // TODO: remove after v0.20
    void saveProject();
};

#endif // MAINWINDOW_H
