/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "midieditorwidget.h"

extern "C" {
#include "ext/midilib/midifile.h"
}

#include "math.h"
#include <QApplication>

#define TextMetaEventHSIZE 3

MidiEditor::MidiEditor(NamedFile &file) : Editor(file), mwidget(new MidiEditorWidget()) {
    QTextDocument *document = mwidget->document();
    QFont font = document->defaultFont();
    font.setFamily("Courier New");
    document->setDefaultFont(font);
    mwidget->setReadOnly(true);
    textType[tMIDI_META::metaTextEvent] = "TEXT";
    textType[tMIDI_META::metaCopyright] = "COPYRIGHT";
    textType[tMIDI_META::metaTrackName] = "TRACK";
    textType[tMIDI_META::metaInstrument] = "INSTRUMENT";
    textType[tMIDI_META::metaLyric] = "LYRIC";
    textType[tMIDI_META::metaMarker] = "MARKER";
    textType[tMIDI_META::metaCuePoint] = "CUE POINT";
}

bool MidiEditor::load() {
    QApplication::setOverrideCursor(Qt::WaitCursor); // show wait cursor
    mwidget->document()->clear();
    MIDI_FILE *midifile = midiFileOpen(path().toStdString().c_str());
    if (!midifile) {
        QApplication::restoreOverrideCursor();
        warning(mwidget, QString("Error opening MIDI file %1").arg(path()));
        return false;
    }
    uint32_t tracks = midiReadGetNumTracks(midifile);
    // show metadata
    int ppq = midiFileGetPPQN(midifile);
    int miditype = midiFileGetVersion(midifile);
    QTextCursor cursor = mwidget->textCursor();
    cursor.insertHtml("<br/>\n<b>File:</b> " + path() + "<br/>\n");
    cursor.insertHtml(QString("<b>MIDI version:</b> %1 ").arg(miditype));
    cursor.insertHtml(QString("<b>PPQ:</b> %1<br/>\n").arg(ppq));
    std::vector<TpmMapEntry> tpmMap;
    MIDI_MSG msg;
    midiReadInitMessage(&msg);
    for (uint32_t t = 0; t < tracks; t++) {
        cursor.insertHtml(QString("<br/>\n<b>Track %1:</b><br/>\n").arg(t));
        double tpm = ppq * 4;
        double measure = 1.0;
        uint32_t lasttick = 0;
        uint32_t mapindex = 0;
        while (midiReadGetNextMessage(midifile, t, &msg)) {
            // MIDI type 1: apply time sig changes from track 0 to all tracks
            if (miditype == 1 && t > 0) {
                while (mapindex < tpmMap.size() && msg.dwAbsPos >= tpmMap[mapindex].tick) {
                    TpmMapEntry entry = tpmMap[mapindex];
                    measure += (entry.tick - lasttick) / tpm;
                    lasttick = entry.tick;
                    tpm = entry.tpm;
                    mapindex++;
                }
                measure += (msg.dwAbsPos - lasttick) / tpm;
                lasttick = msg.dwAbsPos;
            } else {
                measure += msg.dt / tpm;
            }
            int type;
            if (msg.bImpliedMsg) {
                type = msg.iImpliedMsg;
            } else {
                type = msg.iType;
            }
            // print time
            cursor.insertText(QString::asprintf("%07d %09.5f ", msg.dwAbsPos, measure));
            if (type == msgNoteOn) {
                cursor.insertText(QString("[%1] ").arg(msg.MsgData.NoteOn.iChannel - 1));
                cursor.insertText("NOTEON  ");
                int note = msg.MsgData.NoteOn.iNote;
                cursor.insertText(QString("%1 %2").arg(note).arg(notestr(note)));
                cursor.insertText(QString(" vel %1").arg(msg.MsgData.NoteOn.iVolume));
            } else if (type == msgNoteOff) {
                cursor.insertText(QString("[%1] ").arg(msg.MsgData.NoteOff.iChannel - 1));
                cursor.insertText("NOTEOFF ");
                int note = msg.MsgData.NoteOff.iNote;
                cursor.insertText(QString("%1 %2").arg(note).arg(notestr(note)));
                cursor.insertText(QString(" vel %1").arg(msg.MsgData.NoteOff.iVolume));
            } else if (type == msgControlChange) {
                auto cc = msg.MsgData.NoteParameter;
                cursor.insertText(QString("[%1] ").arg(cc.iChannel - 1));
                cursor.insertText(QString("CC %1 value %2").arg(cc.iControl).arg(cc.iParam));
            } else if (type == msgPatchChange) {
                cursor.insertText(QString("[%1] ").arg(msg.MsgData.ChangeProgram.iChannel - 1));
                cursor.insertText(
                    QString("PROGRAM CHANGE %1").arg(msg.MsgData.ChangeProgram.iProgram));
            } else if (type == msgMetaEvent) {
                cursor.insertText("META ");
                tMIDI_META metatype = msg.MsgData.MetaEvent.iType;
                switch (metatype) {
                case tMIDI_META::metaSetTempo:
                    cursor.insertText(
                        QString("Tempo %1").arg(msg.MsgData.MetaEvent.Data.Tempo.fBPM));
                    break;
                case tMIDI_META::metaKeySig: {
                    auto key = msg.MsgData.MetaEvent.Data.KeySig;
                    cursor.insertText(QString("Key Signature %1 %2").arg(key.iKey).arg(key.iMinor));
                    break;
                }
                case tMIDI_META::metaTimeSig: {
                    auto sig = msg.MsgData.MetaEvent.Data.TimeSig;
                    cursor.insertText(QString("Time Signature %1 %2 %3 %4")
                                          .arg(sig.iNom)
                                          .arg(sig.iDenom)
                                          .arg(sig.iClockInMetroTick)
                                          .arg(sig.iNotated32nds));
                    // update ticks per measure
                    tpm = ppq * sig.iNom * 4 / pow(2, sig.iDenom);
                    if (miditype == 1 && t == 0) { // add to tpm map
                        TpmMapEntry entry;
                        entry.tpm = tpm;
                        entry.tick = msg.dwAbsPos;
                        tpmMap.push_back(entry);
                    }
                    break;
                }
                case tMIDI_META::metaMIDIPort:
                    cursor.insertText(
                        QString("MIDI PORT %1").arg(msg.MsgData.MetaEvent.Data.iMIDIPort));
                    break;
                case tMIDI_META::metaSequenceNumber:
                    cursor.insertText(QString("SEQUENCE NUMBER %1")
                                          .arg(msg.MsgData.MetaEvent.Data.iSequenceNumber));
                    break;
                case tMIDI_META::metaEndSequence:
                    cursor.insertText(QString("END SEQUENCE"));
                    break;
                case tMIDI_META::metaSMPTEOffset: {
                    auto data = msg.MsgData.MetaEvent.Data.SMPTE;
                    cursor.insertText(QString("SMTPE offset %1:%2:%3 %4 %5")
                                          .arg(data.iHours)
                                          .arg(data.iMins)
                                          .arg(data.iSecs)
                                          .arg(data.iFrames)
                                          .arg(data.iFF));

                    break;
                }
                case tMIDI_META::metaSequencerSpecific: {
                    cursor.insertText(QString("SEQUENCER SPECIFIC (%1 bytes)")
                                          .arg(msg.iMsgSize - TextMetaEventHSIZE));
                    break;
                }
                case tMIDI_META::metaTextEvent:
                case tMIDI_META::metaCopyright:
                case tMIDI_META::metaTrackName:
                case tMIDI_META::metaInstrument:
                case tMIDI_META::metaLyric:
                case tMIDI_META::metaMarker:
                case tMIDI_META::metaCuePoint: {
                    char *data = (char *)msg.MsgData.MetaEvent.Data.Text.pData;
                    QString body(std::string(data, msg.iMsgSize - TextMetaEventHSIZE).c_str());
                    ;
                    cursor.insertText(QString("%1: %2").arg(textType[metatype]).arg(body));
                    break;
                }
                }
            } else {
                cursor.insertHtml(QString("<i>TBD</i> %1").arg(type));
            }
            cursor.insertText("\n");
        }
    }
    mwidget->moveCursor(QTextCursor::Start);
    QApplication::restoreOverrideCursor();
    return true;
}

QString MidiEditor::notestr(int pitch) {
    // TODO: map names based on key sig
    static QMap<int, QString> notemap = {{0, "C "}, {1, "C#"}, {2, "D "},  {3, "Eb"},
                                         {4, "E "}, {5, "F "}, {6, "F#"},  {7, "G "},
                                         {8, "Ab"}, {9, "A "}, {10, "Bb"}, {11, "B "}};
    return notemap[pitch % 12];
}
