/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef EVENTTABLEMODEL_H
#define EVENTTABLEMODEL_H

#include <QAbstractTableModel>
#include <QDebug>
#include <QTime>

class MidiEvent {
  public:
    int64_t id;
    float measure;
    QString name;
    int32_t arg1;
    int32_t arg2;
    MidiEvent() {}
    MidiEvent(int64_t id) : id(id) {}
    MidiEvent(int64_t id, float measure, QString name, int32_t arg1, int32_t arg2)
        : id(id), measure(measure), name(name), arg1(arg1), arg2(arg2) {}
};

inline bool operator==(const MidiEvent &m1, const MidiEvent &m2) { return m1.id == m2.id; }

class MidiEventTableModel : public QAbstractTableModel {
    Q_OBJECT
  public:
    MidiEventTableModel() : QAbstractTableModel(nullptr) {}
    void add(MidiEvent &event) {
        eventlist.append(event);
        std::sort(eventlist.begin(), eventlist.end(), [](const MidiEvent &e1, const MidiEvent &e2) {
            return (e1.measure < e2.measure);
        });
        emit layoutChanged();
    }
    void remove(uint64_t id) {
        MidiEvent done(id);
        eventlist.removeOne(done);
        emit layoutChanged();
    }
    void clear() {
        eventlist.clear();
        emit layoutChanged();
    }
    // model interface
    int rowCount(const QModelIndex &) const override { return eventlist.length(); }
    int columnCount(const QModelIndex &) const override { return 4; }
    QVariant data(const QModelIndex &index, int role) const override {
        if (role != Qt::DisplayRole) {
            return QVariant::Invalid;
        }
        switch (index.column()) {
        case 0:
            return eventlist[index.row()].measure;
        case 1:
            return eventlist[index.row()].name;
        case 2:
            return eventlist[index.row()].arg1;
        case 3:
            return eventlist[index.row()].arg2;
        default:
            return QVariant();
        }
    }

    QVariant headerData(int section, Qt::Orientation orientation, int role) const override {
        if (role == Qt::DisplayRole) {
            if (orientation == Qt::Horizontal) {
                switch (section) {
                case 0:
                    return "Measure";
                case 1:
                    return "Type";
                case 2:
                    return "Arg 1";
                case 3:
                    return "Arg 2";
                default:
                    return QVariant();
                }

            } else if (orientation == Qt::Vertical) {
                return section + 1;
            }
        }
        return QVariant::Invalid;
    }

  private:
    QList<MidiEvent> eventlist;
};

class ScriptEvent {
  public:
    int64_t id;
    float measure;
    QString name;
    ScriptEvent() {}
    ScriptEvent(int64_t id) : id(id) {}
    ScriptEvent(int64_t id, float measure, QString name) : id(id), measure(measure), name(name) {}
};

inline bool operator==(const ScriptEvent &m1, const ScriptEvent &m2) { return m1.id == m2.id; }

class ScriptEventTableModel : public QAbstractTableModel {
    Q_OBJECT
  public:
    ScriptEventTableModel() : QAbstractTableModel(nullptr) {}
    void add(ScriptEvent &event) {
        eventlist.append(event);
        std::sort(
            eventlist.begin(), eventlist.end(),
            [](const ScriptEvent &e1, const ScriptEvent &e2) { return (e1.measure < e2.measure); });
        emit layoutChanged();
    }
    void remove(uint64_t id) {
        ScriptEvent done(id);
        eventlist.removeOne(done);
        emit layoutChanged();
    }
    void clear() {
        eventlist.clear();
        emit layoutChanged();
    }
    // model interface
    int rowCount(const QModelIndex &) const override { return eventlist.length(); }
    int columnCount(const QModelIndex &) const override { return 2; }
    QVariant data(const QModelIndex &index, int role) const override {
        if (role != Qt::DisplayRole) {
            return QVariant::Invalid;
        }
        const ScriptEvent &row = eventlist[index.row()];
        switch (index.column()) {
        case 0:
            return row.measure;
        case 1:
            return row.name;
        default:
            return QVariant();
        }
    }

    QVariant headerData(int section, Qt::Orientation orientation, int role) const override {
        if (role == Qt::DisplayRole) {
            if (orientation == Qt::Horizontal) {
                switch (section) {
                case 0:
                    return "Measure";
                case 1:
                    return "Name";
                default:
                    return QVariant();
                }

            } else if (orientation == Qt::Vertical) {
                return section + 1;
            }
        }
        return QVariant::Invalid;
    }

  private:
    QList<ScriptEvent> eventlist;
};

class EmptyTableModel : public QAbstractTableModel {
    Q_OBJECT

  public:
    static EmptyTableModel &instance() {
        static EmptyTableModel model;
        return model;
    }
    int rowCount(const QModelIndex &) const override { return 0; }
    int columnCount(const QModelIndex &) const override { return 0; }
    QVariant data(const QModelIndex &, int) const override { return QVariant(); }
};

#endif // EVENTTABLEMODEL_H
