/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "graphnode.h"

#include <QGraphicsSceneMouseEvent>
#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <math.h>

// ------------------- Cable

QRectF Cable::boundingRect() const {
    qreal padding = 2;
    QPointF src = mconnection->external();
    QPointF dest = mconnector->external();
    return QRectF(src, QSizeF(dest.x() - src.x(), dest.y() - src.y()))
        .normalized()
        .adjusted(-padding, -padding, padding, padding);
}

void Cable::deactivate() {
    mactive = false;
    update();
}

Cable::Cable(Connection *connection, Connector *connector, bool audio)
    : mconnection(connection), mconnector(connector), mactive(true), maudio(audio) {
    setZValue(0);
}

void Cable::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *) {
    QPen pen(maudio ? Qt::SolidLine : Qt::DashLine);
    pen.setBrush(mactive ? Qt::black : Qt::gray);
    painter->setPen(pen);
    QLineF line(mconnection->external(), mconnector->external());
    painter->drawLine(line);
}

// ------------------- Clock

Clock::Clock(QString &name, float tempo, uint16_t signum, uint16_t sigden)
    : GraphItem(name), mwidth(80), mheight(65), mactive(true), mmeasure(1.0), mupdated(0),
      mbpm(tempo), msignum(signum), msigden(sigden), mrunning(false) {
    setFlag(ItemIsMovable);
    setFlag(ItemIsSelectable);
    setFlag(ItemSendsGeometryChanges);
    setCacheMode(DeviceCoordinateCache);
    setZValue(1);
}

QRectF Clock::boundingRect() const {
    int padding = 6;
    int32_t w = mwidth + padding;
    int32_t h = mheight + padding;
    return QRectF(-w / 2, -h / 2, w, h);
}

void Clock::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *) {
    QRectF rectangle(-mwidth / 2, -mheight / 2, mwidth, mheight);
    painter->setPen(mactive ? Qt::black : DISABLED);
    painter->drawRoundedRect(rectangle, 27, 27);
    if (name().length() > 0) {
        QPoint textp = QPoint(-19, -15);
        painter->drawText(textp, name());
    }
    // draw measure time
    char time[32];
    sprintf(time, "%07.3f", mmeasure);
    QPoint timep = QPoint(-27, 1);
    painter->setPen(mactive ? Qt::blue : DISABLED);
    painter->drawText(timep, time);
    // beat squares
    float progress = mmeasure - static_cast<uint32_t>(mmeasure);
    int beat = 0;
    for (float i = 0; i < msignum; i++) {
        if (progress > i / msignum) {
            beat = i;
        }
    }
    float stripwidth = 50;
    float beatsize = 5;
    float beatx = -stripwidth / 2;
    const float beaty = 7;
    // painter->drawLine(beatx, beaty + 7, beatx + stripwidth, beaty + 7);
    const float delta = (stripwidth - beatsize) / (msignum - 1);
    painter->setPen(mactive ? Qt::black : DISABLED);
    for (uint16_t i = 0; i < msignum; i++) {
        QRectF beatsquare(beatx, beaty, beatsize, beatsize);
        beatx += delta;
        painter->drawRect(beatsquare);
        if (beat == i) {
            painter->fillRect(beatsquare, mactive ? Qt::green : DISABLED);
        }
    }
    // bpm
    painter->setFont(QFont("Ubuntu", 10));
    QPoint tempop = QPoint(-22, 27);
    sprintf(time, "%.1f", mbpm);
    painter->drawText(tempop, time);
    // play/stop icon
    QPoint iconp = QPoint(14, 18);
    int tside = 8;
    if (mrunning) {
        QPointF out1 = QPointF(iconp.x(), iconp.y());
        QPointF out2 = QPointF(iconp.x(), iconp.y() + tside);
        QPointF out3 = QPointF(iconp.x() + tside, iconp.y() + tside / 2);
        painter->setBrush(mactive ? Qt::green : DISABLED);
        painter->drawPolygon(QPolygonF() << out1 << out2 << out3);
    } else {
        QRectF stop(iconp.x(), iconp.y(), tside, tside);
        painter->fillRect(stop, mactive ? Qt::red : DISABLED);
    }
}

void Clock::updateClock(bool running, float measure, float tempo, uint16_t signum, uint16_t sigden,
                        double now) {
    mrunning = running;
    mmeasure = measure;
    mupdated = now;
    mbpm = tempo;
    msignum = signum;
    msigden = sigden;
    update();
}

void Clock::addScriptEvent(int64_t eid, float measure, QString &name) {
    ScriptEvent evt(eid, measure, name);
    scriptEventModel.add(evt);
}

void Clock::time(float measure, double now) {
    mmeasure = measure;
    mupdated = now;
    update();
}

void Clock::tick(double now) {
    double delta = (now - mupdated) * mbpm * msigden / (msignum * 4.0);
    if (mrunning) {
        mmeasure += delta;
        mupdated = now;
        update();
    }
}

void Clock::deactivate() {
    mactive = false;
    update();
}

// ------------------- GraphNode

GraphNode::GraphNode(QString &name, uint16_t inputs, uint16_t outputs, uint16_t midiins,
                     uint16_t midiouts)
    : GraphItem(name), minputs(inputs), moutputs(outputs), mmidiins(midiins), mmidiouts(midiouts),
      mactive(true) {
    uint16_t max = allins() > allouts() ? allins() : allouts();
    mheight = max ? max * PORT_SPACING : 100;
    QFont font;
    QFontMetrics fm(font);
    fonth = fm.height();
    int startwidth = fm.horizontalAdvance(name) + 15;
    if (moutputs) {
        startwidth += METER_SPACE;
    } else if (mmidiouts) {
        startwidth += MIDI_METER_SPACE;
    }
    width(startwidth);
    maudiolevel.assign(outputs, 0.0);
    mmidilevel.assign(midiouts, false);
    setFlag(ItemIsMovable);
    setFlag(ItemIsSelectable);
    setFlag(ItemSendsGeometryChanges);
    setCacheMode(DeviceCoordinateCache);
    setZValue(1);
}

QRectF GraphNode::boundingRect() const {
    int padding = 6;
    int32_t w = mwidth + padding;
    int32_t h = mheight + padding;
    return QRectF(-w / 2, -h / 2, w, h);
}

float ratio2db(float ratio) { return 20 * log10(ratio); }

void GraphNode::drawMeter(QPainter *painter, float ratio, QPointF start) {
    QColor unlit = Qt::lightGray;
    QColor lit = mactive ? Qt::green : Qt::darkGray;
    QColor high = mactive ? Qt::yellow : Qt::darkGray;
    QColor clip = mactive ? Qt::red : Qt::darkGray;
    QColor lights[7];
    if (ratio > 0) {
        float db = ratio2db(ratio);
        lights[0] = lit;
        for (int j = 1; j <= 4; j++) {
            float threshhold = -60 + j * 10;
            lights[j] = db > threshhold ? lit : unlit;
        }
        lights[5] = db > -10 ? high : unlit;
        lights[6] = ratio >= 1.0 ? clip : unlit;
    } else {
        for (int i = 0; i < 7; i++) {
            lights[i] = unlit;
        }
    }
    QColor divider = Qt::white;
    int miniw = 2; // for mini bars on either end
    int cellw = 5;
    int cellh = 8;
    int divw = 1;
    float x = start.x();
    float y = start.y() - cellh / 2;
    painter->fillRect(x, y, miniw, cellh, lights[0]);
    x += miniw;
    painter->fillRect(x, y, divw, cellh, divider);
    x += divw;
    for (int i = 1; i <= 5; i++) {
        painter->fillRect(x, y, cellw, cellh, lights[i]);
        x += cellw;
        painter->fillRect(x, y, divw, cellh, divider);
        x += divw;
    }
    painter->fillRect(x, y, miniw, cellh, lights[6]);
}

void GraphNode::drawMidiMeter(QPainter *painter, float activity, QPointF start) {
    QColor unlit = Qt::lightGray;
    QColor lit = mactive ? Qt::blue : Qt::darkGray;
    // QColor clip = mactive ? Qt::red : Qt::darkGray;

    QColor color1 = activity ? lit : unlit;
    QColor color2 = unlit; // TODO: implement clipping

    QColor divider = Qt::white;
    int miniw = 2; // for mini bars on end
    int cellw = 5;
    int cellh = 8;
    int divw = 1;
    float x = start.x();
    float y = start.y() - cellh / 2;
    painter->fillRect(x, y, cellw, cellh, color1);
    x += cellw;
    painter->fillRect(x, y, divw, cellh, divider);
    x += divw;
    painter->fillRect(x, y, miniw, cellh, color2);
}

void GraphNode::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *) {
    QPen blackpen;
    blackpen.setBrush(mactive ? Qt::black : DISABLED);
    // draw outer rectangle
    QPen bluepen(Qt::blue);
    painter->setPen(selected() ? bluepen : blackpen);
    QRectF rectangle(-mwidth / 2, -mheight / 2, mwidth, mheight);
    painter->drawRoundedRect(rectangle, 7, 7);
    // draw ports
    for (int i = 0; i < allins(); i++) {
        Connector *port = mconnector[i];
        painter->fillRect(port->rectangle(), port->audio() ? Qt::white : blackpen.color());
        painter->drawRect(port->rectangle());
    }
    for (uint16_t i = 0; i < allouts(); i++) {
        Connection *port = mconnection[i];
        painter->fillRect(port->rectangle(), port->audio() ? Qt::white : blackpen.color());
        painter->drawRect(mconnection[i]->rectangle());
        QPointF point = mconnection[i]->internal();
        // connection meter
        if (port->audio()) {
            QPointF meterloc = QPointF(point.x() - METER_SPACE, point.y());
            drawMeter(painter, maudiolevel[i - mmidiouts], meterloc);
            //        snprintf(nstr, 128, "%f", maudiolevel[i]);
            //        painter->drawText(moutput[i]->meterloc(), nstr);
        } else {
            QPointF meterloc = QPointF(point.x() - MIDI_METER_SPACE, point.y());
            drawMidiMeter(painter, mmidilevel[i], meterloc);
        }
    }
    // paint name
    if (name().length() > 0) {
        QPoint textp = QPoint(8 - mwidth / 2, fonth / 3);
        painter->drawText(textp, name());
    }
}

QVariant GraphNode::itemChange(GraphicsItemChange change, const QVariant &value) {
    switch (change) {
    case ItemPositionHasChanged:
        for (int i = 0; i < allins(); i++) {
            Cable *cable = mconnector[i]->cable();
            if (cable) {
                cable->moved();
            }
        }
        for (int i = 0; i < allouts(); i++) {
            for (Cable *cable : mconnection[i]->cables()) {
                cable->moved();
            }
        }
        break;
    default:
        break;
    };

    return QGraphicsItem::itemChange(change, value);
}

int16_t GraphNode::level() {
    if (mlevel == -1) { // calculate level
        int max = 0;
        // start with level 1 if connected at all
        for (uint16_t i = 0; i < allouts(); i++) {
            if (mconnection[i]->cables().size()) {
                max = 1;
                break;
            }
        }
        // level is 1 plus largest input level
        for (uint16_t i = 0; i < allins(); i++) {
            if (mconnector[i]->cable()) {
                GraphNode *child = mconnector[i]->cable()->connection()->parent();
                int childlevel = 1 + child->level();
                if (childlevel > max) {
                    max = childlevel;
                }
            }
        }
        mlevel = max;
    }
    return mlevel;
}

void GraphNode::addProcessorMidi(uint16_t midiins, uint16_t midiouts) {
    mmidiins = midiins;
    mmidiouts = midiouts;
    uint16_t ins = allins();
    uint16_t outs = allouts();
    int16_t x = mwidth / 2;
    if (ins) {
        int16_t start = ((1 - ins) * PORT_SPACING) / 2;
        Connector **oldport = mconnector;
        mconnector = new Connector *[ins];
        for (int i = 0; i < ins; i++) {
            if (i < midiins) {
                mconnector[i] = new Connector(this, QString("MIDI input %1").arg(i + 1), -x,
                                              start + i * PORT_SPACING, false);
            } else {
                mconnector[i] = oldport[i - midiins];
                mconnector[i]->y(start + i * PORT_SPACING);
            }
        }
    }
    if (outs) {
        int16_t start = ((1 - outs) * PORT_SPACING) / 2;
        mconnection = new Connection *[outs];
        for (int i = 0; i < outs; i++) {
            mconnection[i] = new Connection(this, x, start + i * PORT_SPACING, i >= mmidiouts);
        }
    }
}

void GraphNode::audiolevels(float level[]) {
    maudiolevel.assign(level, level + moutputs);
    update();
}

void GraphNode::midilevels(uint16_t index, bool activity) {
    mmidilevel[index] = activity;
    update();
}

void GraphNode::deactivate() {
    mactive = false;
    for (int i = 0; i < allouts(); i++) {
        for (Cable *cable : mconnection[i]->cables()) {
            cable->deactivate();
        }
    }
    update();
}

QList<QString> GraphNode::eventConnectors() {
    QList<QString> ret;
    for (uint16_t i = 0; i < midiins(); i++) {
        ret.append(mconnector[i]->name());
    }
    return ret;
}

QList<Clock *> GraphNode::eventClocks(QString cname) {
    return connectorByName(cname)->eventClocks();
}

void GraphNode::removeClock(Clock *clock) {
    for (int i = 0; i < allins(); i++) {
        input(i)->removeClock(clock);
    }
}

Connector *GraphNode::connectorByName(QString name) {
    for (int i = 0; i < midiins(); i++) {
        if (mconnector[i]->name() == name) {
            return mconnector[i];
        }
    }
    throw std::logic_error("No connector found for name");
}

void GraphNode::createPorts() {
    uint16_t ins = allins();
    uint16_t outs = allouts();
    int16_t x = mwidth / 2;
    if (ins) {
        int16_t start = ((1 - ins) * PORT_SPACING) / 2;
        mconnector = new Connector *[ins];
        for (int i = 0; i < ins; i++) {
            mconnector[i] = new Connector(this, QString("MIDI input %1").arg(i + 1), -x,
                                          start + i * PORT_SPACING, i >= mmidiins);
        }
    }
    if (outs) {
        int16_t start = ((1 - outs) * PORT_SPACING) / 2;
        mconnection = new Connection *[outs];
        for (int i = 0; i < outs; i++) {
            mconnection[i] = new Connection(this, x, start + i * PORT_SPACING, i >= mmidiouts);
        }
    }
}

void GraphNode::width(uint16_t width) {
    mwidth = width;
    mwidth += mwidth % 2;
}

MixerNode::MixerNode(QString &name, uint16_t inputs, uint16_t outputs)
    : GraphNode(name, inputs, outputs, 0, 0), selectin(-1), selectout(-1) {
    mgain = new float *[inputs];
    wires = new QPainterPath *[inputs];
    width(height() / 2 + METER_SPACE);
    createPorts();
    for (int i = 0; i < inputs; i++) {
        mgain[i] = new float[outputs];
        std::fill(mgain[i], mgain[i] + outputs, 0.0);
        wires[i] = new QPainterPath[outputs];
        for (int j = 0; j < outputs; j++) {
            // find source and destination points
            QPointF source = input(i)->internal();
            QPointF con = output(j)->internal();
            QPointF dest = QPointF(con.x() - METER_SPACE, con.y());
            // calculate control points
            float delta = abs(source.x() - dest.x()) / 2;
            QPointF c1(source.x() + delta, source.y());
            QPointF c2(dest.x() - delta, dest.y());
            // create cubic bezier
            wires[i][j] = QPainterPath(source);
            wires[i][j].cubicTo(c1, c2, dest);
        }
    }
    mwirecolor.append(Qt::black);
    mwirecolor.append(Qt::darkRed);
    mwirecolor.append(Qt::darkYellow);
    mwirecolor.append(Qt::blue);
    mwirecolor.append(Qt::green);
    mwirecolor.append(Qt::cyan);
    mwirecolor.append(Qt::magenta);
    mwirecolor.append(Qt::red);
}

void MixerNode::gain(int input, int output, float gain) {
    mgain[input][output] = gain;
    selectin = input;
    selectout = output;
    update();
}

void MixerNode::gains(int input, float *gains) {
    for (int o = 0; o < outputs(); o++) {
        mgain[input][o] = gains[o];
    }
    update();
}

void MixerNode::mouseReleaseEvent(QGraphicsSceneMouseEvent *event) {
    QGraphicsItem::mouseReleaseEvent(event);
    QPointF mouse = event->pos();
    selectin = -1;
    selectout = -1;
    float min = 100; // 10 pixels squared
    // find minimum distance to any wire
    for (int i = 0; i < inputs(); i++) {
        for (int j = 0; j < outputs(); j++) {
            if (mgain[i][j]) {
                for (float p = 0.1; p < 1.00; p += 0.1) {
                    QPointF wire = wires[i][j].pointAtPercent(p);
                    float dx = mouse.x() - wire.x();
                    float dy = mouse.y() - wire.y();
                    float length2 = dx * dx + dy * dy;
                    if (length2 < min) {
                        selectin = i;
                        selectout = j;
                        min = length2;
                    }
                }
            }
        }
    }
    update();
}

void MixerNode::paint(QPainter *painter, const QStyleOptionGraphicsItem *item, QWidget *widget) {
    GraphNode::paint(painter, item, widget);
    // draw gain lines
    for (int i = 0; i < inputs(); i++) {
        for (int j = 0; j < outputs(); j++) {
            if (mgain[i][j] > 0.0) {
                // set pen
                QPen pen;
                pen.setWidth(mgain[i][j]);
                pen.setBrush(wirecolor(j));
                painter->setPen(pen);
                painter->drawPath(wires[i][j]);
            }
        }
    }
    if (selectin > -1) {
        // set pen
        QPen pen;
        pen.setWidth(1);
        pen.setBrush(wirecolor(selectout));
        painter->setPen(pen);
        // set font
        QFont font("Ubuntu", 8);
        QFontMetrics fmetrics(font);
        painter->setFont(font);
        // gain amount as string
        char numstr[128];
        snprintf(numstr, 28, "%.2f", mgain[selectin][selectout]);
        QPointF inp = input(selectin)->internal();
        int padding = 3;
        QPointF textp = QPointF(inp.x() + padding, inp.y() + fmetrics.descent());
        QRectF bsquare = fmetrics.boundingRect(numstr);
        bsquare.translate(textp.x(), textp.y());
        // draw things
        painter->fillRect(bsquare, Qt::white);                           // backing squre
        painter->drawRoundedRect(bsquare + QMarginsF(1, 0, 1, 0), 2, 2); // outline box
        painter->drawText(textp, numstr);
        painter->fillRect(output(selectout)->rectangle(), wirecolor(selectout)); // output port
    }
}

Lv2Node::Lv2Node(QString &name, uint16_t inputs, uint16_t outputs, uint16_t midiins,
                 uint16_t midiouts)
    : GraphNode(name, inputs, outputs, midiins, midiouts) {
    createPorts();
}

void Lv2Node::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) {
    GraphNode::paint(painter, option, widget);
}

QRectF InputNode::boundingRect() const {
    QRectF grect = GraphNode::boundingRect();
    return QRectF(grect.x() - TRIANGLE_SIZE, grect.y(), grect.width() + TRIANGLE_SIZE,
                  grect.height());
}

void InputNode::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) {
    GraphNode::paint(painter, option, widget);
    // input triangles
    for (int i = 0; i < allouts(); i++) {
        QPointF out = output(i)->internal();
        QPointF in1 = QPointF(-width() / 2, out.y() - TRIANGLE_SIZE / 2);
        QPointF in2 = QPointF(-width() / 2, out.y() + TRIANGLE_SIZE / 2);
        QPointF in3 = QPointF(-width() / 2 - TRIANGLE_SIZE, out.y());
        painter->setBrush(midiouts() ? active() ? Qt::black : DISABLED : Qt::white);
        painter->drawPolygon(QPolygonF() << in1 << in2 << in3);
    }
}

QRectF OutputNode::boundingRect() const {
    QRectF grect = GraphNode::boundingRect();
    return QRectF(grect.x(), grect.y(), grect.width() + TRIANGLE_SIZE, grect.height());
}

void OutputNode::audiolevels(float level[]) {
    maudiolevel.assign(level, level + inputs());
    update();
}

void OutputNode::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) {
    GraphNode::paint(painter, option, widget);
    // output triangles
    for (int i = 0; i < allins(); i++) {
        QPointF in = input(i)->internal();
        QPointF out1 = QPointF(width() / 2, in.y() - TRIANGLE_SIZE / 2);
        QPointF out2 = QPointF(width() / 2, in.y() + TRIANGLE_SIZE / 2);
        QPointF out3 = QPointF(width() / 2 + TRIANGLE_SIZE, in.y());
        painter->setBrush(midiins() ? active() ? Qt::black : DISABLED : Qt::white);
        painter->drawPolygon(QPolygonF() << out1 << out2 << out3);
        // meters
        if (inputs()) {
            QPointF meterloc = QPointF(width() / 2 - METER_SPACE, in.y());
            drawMeter(painter, maudiolevel[i], meterloc);
        } else {
            QPointF meterloc = QPointF(width() / 2 - MIDI_METER_SPACE, in.y());
            drawMidiMeter(painter, mmidilevel[i], meterloc);
        }
    }
}

QPointF Connection::external() { return QPointF(mx + PSIZE / 2, my) + mparent->pos(); }

QPointF Connector::external() { return QPointF(mx - PSIZE / 2, my) + mparent->pos(); }

void Connector::addMidiEvent(Clock *clock, MidiEvent &evt) { midimodel(clock)->add(evt); }

void Connector::removeMidiEvent(Clock *clock, int64_t eid) { midimodel(clock)->remove(eid); }

void Connector::clearMidiEvents(Clock *clock) { midimodel(clock)->clear(); }

MidiEventTableModel *Connector::midimodel(Clock *clock) {
    if (!tableModel.count(clock)) {
        tableModel[clock] = new MidiEventTableModel;
    }
    return tableModel[clock];
}
