/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef GRAPHNODE_H
#define GRAPHNODE_H

#include <QGraphicsItem>
#include <QSet>

#include "eventtablemodel.h"

class Cable;
class GraphNode;

class Port {
  protected:
    GraphNode *mparent;
    int mx;
    int my;
    const static int PSIZE = 6;
    bool maudio;

  public:
    Port(GraphNode *parent, int x, int y, bool audio)
        : mparent(parent), mx(x), my(y), maudio(audio) {}
    QRectF rectangle() { return QRectF(mx - PSIZE / 2, my - PSIZE / 2, PSIZE, PSIZE); }
    bool audio() { return maudio; }
    void y(int y) { my = y; }
};

class Connection : public Port {
    QSet<Cable *> mcables;

  public:
    Connection(GraphNode *parent, int x, int y, bool audio) : Port(parent, x, y, audio) {}
    QPoint internal() { return QPoint(mx - PSIZE / 2, my); }
    QPointF external();
    GraphNode *parent() { return mparent; }
    void connect(Cable *cable) { mcables.insert(cable); }
    void disconnect(Cable *cable) { mcables.remove(cable); }
    QSet<Cable *> cables() { return mcables; }
    void remove(Cable *cable) { mcables.remove(cable); }
};

const QColor DISABLED = Qt::gray;

class Clock;
class Connector;

// superclass for all selectable nodes
class GraphItem : public QGraphicsItem {
    QString mname;
    bool mselected;

  public:
    GraphItem(QString &name) : mname(name), mselected(false) {}
    QString name() { return mname; }
    bool selected() { return mselected; }
    void selected(bool selected) { mselected = selected; }
    virtual QList<QString> eventConnectors() { return QList<QString>(); }
    virtual QList<Clock *> eventClocks(QString) { return QList<Clock *>(); }
    virtual QAbstractTableModel *eventModel(int connectorindex, int clockindex) = 0;
};

class Clock : public GraphItem { // TODO: split into Clock + ClockNode
    uint16_t mwidth;
    uint16_t mheight;
    bool mactive;
    // clock data
    float mmeasure;
    double mupdated; // when was mmeasure updated
    float mbpm;
    uint32_t msignum;
    uint32_t msigden;
    bool mrunning;
    ScriptEventTableModel scriptEventModel;

  public:
    Clock(QString &name, float tempo, uint16_t signum, uint16_t sigden);
    QRectF boundingRect() const override;
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *) override;
    uint16_t width() { return mwidth; }
    void updateClock(bool running, float measure, float tempo, uint16_t signum, uint16_t sigden,
                     double now);
    void addScriptEvent(int64_t eid, float measure, QString &name);
    void removeScriptEvent(int64_t eid) { scriptEventModel.remove(eid); }
    void clearScriptEvents() { scriptEventModel.clear(); }
    void time(float measure, double now);
    void tick(double now);
    bool active() { return mactive; }
    void deactivate();
    QList<QString> eventConnectors() override { return {"Script Events"}; }
    QList<Clock *> eventClocks(QString) override { return {this}; }
    QAbstractTableModel *eventModel(int, int) override { return &scriptEventModel; }
};

class Connector : public Port {
    Cable *mcable;
    QString mname;
    QMap<Clock *, MidiEventTableModel *> tableModel;

  public:
    Connector(GraphNode *parent, QString name, int x, int y, bool audio)
        : Port(parent, x, y, audio), mcable(nullptr), mname(name) {}
    QString name() { return mname; }
    QPoint internal() { return QPoint(mx + PSIZE / 2, my); }
    QPointF external();
    void connect(Cable *cable) { mcable = cable; }
    void disconnect() { mcable = nullptr; }
    Cable *cable() { return mcable; }
    void addMidiEvent(Clock *clock, MidiEvent &event);
    void removeMidiEvent(Clock *clock, int64_t eid);
    void clearMidiEvents(Clock *clock);
    QList<Clock *> eventClocks() { return tableModel.keys(); }
    void removeClock(Clock *clock) {
        delete tableModel[clock];
        tableModel.remove(clock);
    }
    QAbstractTableModel *eventModel(Clock *clock) { return tableModel[clock]; }

  private:
    MidiEventTableModel *midimodel(Clock *clock);
};

class Cable : public QGraphicsItem {
    Connection *mconnection;
    Connector *mconnector;
    bool mactive;
    bool maudio;

  public:
    Cable(Connection *connection, Connector *connector, bool audio);
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *);
    QRectF boundingRect() const;
    void moved() { prepareGeometryChange(); }
    Connection *connection() { return mconnection; }
    Connector *connector() const { return mconnector; }
    void deactivate();
};

class GraphNode : public GraphItem {
    uint16_t mheight;
    uint16_t mwidth;
    Connection **mconnection;
    Connector **mconnector;
    uint16_t minputs;
    uint16_t moutputs;
    uint16_t mmidiins;
    uint16_t mmidiouts;
    int16_t mlevel;
    const int PORT_SPACING = 22;
    bool mactive;

  protected:
    const int METER_SPACE = 37;
    const int MIDI_METER_SPACE = 10;
    std::vector<float> maudiolevel;
    std::vector<bool> mmidilevel;
    int fonth;
    void createPorts();
    void width(uint16_t width);
    bool active() { return mactive; }
    void drawMeter(QPainter *painter, float ratio, QPointF start);
    void drawMidiMeter(QPainter *painter, float ratio, QPointF start);

  public:
    GraphNode(QString &name, uint16_t inputs, uint16_t outputs, uint16_t midiins,
              uint16_t midiouts);
    QRectF boundingRect() const override;
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *) override;
    QVariant itemChange(GraphicsItemChange change, const QVariant &value) override;
    uint16_t width() { return mwidth; }
    uint16_t height() { return mheight; }
    uint16_t inputs() { return minputs; }
    uint16_t outputs() { return moutputs; }
    uint16_t midiins() { return mmidiins; }
    uint16_t midiouts() { return mmidiouts; }
    uint16_t allins() { return minputs + mmidiins; }
    uint16_t allouts() { return moutputs + mmidiouts; }
    Connector *input(uint16_t i) { return mconnector[i]; }
    Connection *output(uint16_t i) { return mconnection[i]; }
    void level(int16_t level) { mlevel = level; }
    int16_t level();
    void addProcessorMidi(uint16_t ins, uint16_t outs);
    virtual void audiolevels(float level[]);
    virtual void midilevels(uint16_t index, bool activity);
    void deactivate();
    QList<QString> eventConnectors() override;
    QList<Clock *> eventClocks(QString cname) override;
    void removeClock(Clock *clock);
    QAbstractTableModel *eventModel(int connectoridx, int clockidx) override {
        auto connector = mconnector[connectoridx];
        return connector->eventModel(connector->eventClocks()[clockidx]);
    }

  private:
    Connector *connectorByName(QString name);
};

class MixerNode : public GraphNode {
    float **mgain;
    QPainterPath **wires;
    QList<QColor> mwirecolor;
    int selectin;
    int selectout;

  public:
    MixerNode(QString &name, uint16_t inputs, uint16_t outputs);
    void gain(int input, int output, float gain);
    void gains(int input, float *gains);
    // for graphics item
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) override;
    void mouseReleaseEvent(QGraphicsSceneMouseEvent *event) override;

  private:
    QColor wirecolor(int i) { return active() ? mwirecolor[i % mwirecolor.size()] : DISABLED; }
};

class Lv2Node : public GraphNode {
  public:
    Lv2Node(QString &name, uint16_t inputs, uint16_t outputs, uint16_t midiins, uint16_t midiouts);
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);
};

class InputNode : public GraphNode {
    const float TRIANGLE_SIZE = 6;

  public:
    InputNode(QString &name, uint16_t channels, uint16_t midichannels)
        : GraphNode(name, 0, channels, 0, midichannels) {
        createPorts();
    }
    QRectF boundingRect() const override;
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) override;
};

class OutputNode : public GraphNode {
    const float TRIANGLE_SIZE = 6;

  public:
    OutputNode(QString &name, uint16_t channels, uint16_t midichannels)
        : GraphNode(name, channels, 0, midichannels, 0) {
        width(width() + (channels ? METER_SPACE : MIDI_METER_SPACE));
        maudiolevel.assign(channels, 0.0);
        mmidilevel.assign(midichannels, false);
        createPorts();
    }
    QRectF boundingRect() const override;
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget) override;
    void audiolevels(float level[]) override;
};

#endif // GRAPHNODE_H
