#include "objectgraph.h"

#include <sys/time.h>

ObjectGraph::ObjectGraph(QWidget *parent) : QGraphicsView(parent), mselected(nullptr) {
    scene = new QGraphicsScene(this);
    scene->setItemIndexMethod(QGraphicsScene::NoIndex);
    setScene(scene);
    setCacheMode(CacheBackground);
    setViewportUpdateMode(BoundingRectViewportUpdate);
    setRenderHint(QPainter::Antialiasing);
    setTransformationAnchor(AnchorUnderMouse);
    setWindowTitle(tr("Object Graph"));
    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &ObjectGraph::clockTick);
    connect(scene, &QGraphicsScene::selectionChanged, this, &ObjectGraph::onSelection);
}

void ObjectGraph::cconnect(int64_t port1, int64_t port2) {
    Connection *connection = connectionMap[port1];
    Connector *connector = connectorMap[port2];
    if (connection && connector) {
        // remove any existing cable on connector
        Cable *existing = connector->cable();
        if (existing) {
            existing->connection()->remove(existing);
            scene->removeItem(existing);
            delete existing;
        }
        Cable *cable = new Cable(connection, connector, connector->audio());
        connection->connect(cable);
        connector->connect(cable);
        scene->addItem(cable);
        layout();
    } else {
        printf("Couldn't connect %p -> %p\n", connection, connector);
    }
}

void ObjectGraph::addClock(int64_t hash, QString &name, float tempo, uint16_t signum,
                           uint16_t sigden) {
    Clock *clock = new Clock(name, tempo, signum, sigden);
    // start timer if this is the first clock
    if (clockMap.size() == 0) {
        timer->start(100); // ms
    }
    clockMap[hash] = clock;
    scene->addItem(clock);
    layout();
}

void ObjectGraph::addLv2Plugin(int64_t hash, QString &name, int inputs, int outputs, int midiins,
                               int midiouts) {
    addNode(new Lv2Node(name, inputs, outputs, midiins, midiouts), hash);
}

void ObjectGraph::addProcessorMidi(int64_t hash, int midiins, int midiouts) {
    if (nodeMap.count(hash)) {
        GraphNode *node = nodeMap[hash];
        node->addProcessorMidi(midiins, midiouts);
    }
}

void ObjectGraph::addMidiEvent(int64_t port, int64_t clockid, int64_t eid, float measure,
                               QString &name, int arg1, int arg2) {
    Connector *connector = connectorMap[port];
    if (connector) {
        if (clockMap.count(clockid)) {
            Clock *clock = static_cast<Clock *>(clockMap[clockid]);
            MidiEvent evt(eid, measure, name, arg1, arg2);
            connector->addMidiEvent(clock, evt);
        }
    } else {
        printf("midi event unknown connector %p\n", connector);
    }
}

void ObjectGraph::addScriptEvent(int64_t node, int64_t eid, float measure, QString &name) {
    if (clockMap.count(node)) {
        Clock *clock = static_cast<Clock *>(clockMap[node]);
        clock->addScriptEvent(eid, measure, name);
    }
}

void ObjectGraph::removeMidiEvent(int64_t port, int64_t clockid, int64_t event) {
    Connector *connector = connectorMap[port];
    if (connector) {
        if (clockMap.count(clockid)) {
            Clock *clock = static_cast<Clock *>(clockMap[clockid]);
            connector->removeMidiEvent(clock, event);
        }
    } else {
        printf("midi remove event unknown connector %p\n", connector);
    }
}

void ObjectGraph::clearMidiEvents(int64_t port, int64_t clockid) {
    Connector *connector = connectorMap[port];
    if (connector) {
        if (clockMap.count(clockid)) {
            Clock *clock = static_cast<Clock *>(clockMap[clockid]);
            connector->clearMidiEvents(clock);
        }
    } else {
        printf("midi clear event unknown connector %p\n", connector);
    }
}

void ObjectGraph::removeScriptEvent(int64_t node, int64_t eid) {
    if (clockMap.count(node)) {
        Clock *clock = static_cast<Clock *>(clockMap[node]);
        clock->removeScriptEvent(eid);
    }
}

void ObjectGraph::clearScriptEvents(int64_t node) {
    if (clockMap.count(node)) {
        Clock *clock = static_cast<Clock *>(clockMap[node]);
        clock->clearScriptEvents();
    }
}

void ObjectGraph::addMixer(int64_t mixer, QString &name, int inputs, int outputs) {
    addNode(new MixerNode(name, inputs, outputs), mixer);
}

void ObjectGraph::addInput(int64_t hash, QString &name, int channels) {
    addNode(new InputNode(name, channels, 0), hash);
}

void ObjectGraph::addOutput(int64_t hash, QString &name, int channels) {
    addNode(new OutputNode(name, channels, 0), hash);
}

void ObjectGraph::addMidiInput(int64_t hash, QString &name, int channels) {
    addNode(new InputNode(name, 0, channels), hash);
}

void ObjectGraph::addMidiOutput(int64_t hash, QString &name, int channels) {
    addNode(new OutputNode(name, 0, channels), hash);
}

void ObjectGraph::mixerGain(int64_t mixer, int input, int output, float gain) {
    if (nodeMap.count(mixer)) {
        MixerNode *node = static_cast<MixerNode *>(nodeMap[mixer]);
        node->gain(input, output, gain);
    } else {
        fprintf(stderr, "mixerGain() unknown node %ld\n", mixer);
    }
}

void ObjectGraph::mixerGains(int64_t mixer, int input, float *gains) {
    if (nodeMap.count(mixer)) {
        MixerNode *node = static_cast<MixerNode *>(nodeMap[mixer]);
        node->gains(input, gains);
    } else {
        fprintf(stderr, "mixerGains() unknown node %ld\n", mixer);
    }
}

void ObjectGraph::clockTime(int64_t node, float measure) {
    if (clockMap.count(node)) {
        Clock *clock = static_cast<Clock *>(clockMap[node]);
        clock->time(measure, timenow());
    }
}

void ObjectGraph::clockUpdate(int64_t node, bool running, float measure, float tempo,
                              uint16_t signum, uint16_t sigden) {
    if (clockMap.count(node)) {
        // update this clock display
        Clock *clock = static_cast<Clock *>(clockMap[node]);
        clock->updateClock(running, measure, tempo, signum, sigden, timenow());
        // TODO: send update to all event lists
    }
}

void ObjectGraph::onSelection() {
    QList<QGraphicsItem *> items = scene->selectedItems();
    if (mselected) {
        mselected->selected(false);
    }
    if (items.size()) {
        mselected = static_cast<GraphItem *>(items[0]);
        mselected->selected(true);
        emit selected(mselected);
    } else {
        emit selected(nullptr);
    }
}

void ObjectGraph::nodeInputs(int64_t nodeid, int64_t *ports) {
    if (nodeMap.count(nodeid)) {
        GraphNode *node = nodeMap[nodeid];
        uint16_t start = node->midiins();
        for (uint16_t i = 0; i < node->inputs(); i++) {
            connectorMap[ports[i]] = node->input(start + i);
        }
    } else {
        fprintf(stderr, "nodeInputs() unknown node %ld\n", nodeid);
    }
}

void ObjectGraph::nodeOutputs(int64_t nodeid, int64_t *ports) {
    if (nodeMap.count(nodeid)) {
        GraphNode *node = nodeMap[nodeid];
        uint16_t start = node->midiouts();
        for (uint16_t i = 0; i < node->outputs(); i++) {
            connectionMap[ports[i]] = node->output(start + i);
        }
    } else {
        fprintf(stderr, "nodeOutputs() unknown node %ld\n", nodeid);
    }
}

void ObjectGraph::nodeMidiInputs(int64_t nodeid, int64_t *ports) {
    if (nodeMap.count(nodeid)) {
        GraphNode *node = nodeMap[nodeid];
        for (uint16_t i = 0; i < node->midiins(); i++) {
            connectorMap[ports[i]] = node->input(i);
        }
    } else {
        fprintf(stderr, "nodeMidiInputs() unknown node %ld\n", nodeid);
    }
}

void ObjectGraph::nodeMidiOutputs(int64_t nodeid, int64_t *ports) {
    if (nodeMap.count(nodeid)) {
        GraphNode *node = nodeMap[nodeid];
        for (uint16_t i = 0; i < node->midiouts(); i++) {
            connectionMap[ports[i]] = node->output(i);
        }
    } else {
        fprintf(stderr, "nodeMidiOutputs() unknown node %ld\n", nodeid);
    }
}

void ObjectGraph::levels(int64_t nodeid, float levels[]) {
    if (nodeMap.count(nodeid)) {
        nodeMap[nodeid]->audiolevels(levels);
    }
}

void ObjectGraph::midilevels(int64_t nodeid, uint16_t index, float level) {
    if (nodeMap.count(nodeid)) {
        nodeMap[nodeid]->midilevels(index, level);
    }
}

void ObjectGraph::completed() {
    for (const auto &pair : nodeMap) {
        GraphNode *node = pair.second;
        node->deactivate();
    }
    for (const auto &pair : clockMap) {
        Clock *node = pair.second;
        node->deactivate();
    }
    timer->stop();
}

double ObjectGraph::timenow() {
    struct timeval nowtv {};
    gettimeofday(&nowtv, nullptr);
    // time in fractional minutes
    return (nowtv.tv_sec / 60.0) + (nowtv.tv_usec / 60000000.0);
}

void ObjectGraph::clockTick() {
    double now = timenow();
    for (const auto &pair : clockMap) {
        Clock *node = pair.second;
        if (node->active()) {
            node->tick(now);
        }
    }
}

void ObjectGraph::addNode(GraphNode *node, int64_t hash) {
    nodeMap[hash] = node;
    scene->addItem(node);
    layout();
}

void ObjectGraph::finishClock(int64_t clockid) {
    if (clockMap.count(clockid)) {
        Clock *clock = static_cast<Clock *>(clockMap[clockid]);
        clock->deactivate();
    }
}

void ObjectGraph::removeNode(int64_t nodeid) {
    GraphNode *node = nodeMap[nodeid];
    for (uint16_t i = 0; i < node->allins(); i++) {
        Connector *inport = node->input(i);
        Cable *cable = inport->cable();
        if (cable) {
            Connection *outport = cable->connection();
            outport->disconnect(cable);
            scene->removeItem(cable);
            delete cable;
        }
    }
    for (uint16_t i = 0; i < node->allouts(); i++) {
        Connection *outport = node->output(i);
        for (Cable *cable : outport->cables()) {
            Connector *inport = cable->connector();
            inport->disconnect();
            scene->removeItem(cable);
            delete cable;
        }
    }
    nodeMap.erase(nodeid);
    scene->removeItem(node);
    delete node;
}

void ObjectGraph::scriptStart() {
    // this may not be the first run - remove old clocks
    auto pair = clockMap.cbegin();
    while (pair != clockMap.cend()) {
        Clock *clock = pair->second;
        if (!clock->active()) {
            scene->removeItem(clock);
            for (const auto &pair : nodeMap) {
                GraphNode *node = pair.second;
                node->removeClock(clock);
            }
            delete clock;
            pair = clockMap.erase(pair);
        } else {
            ++pair;
        }
    }
}

void ObjectGraph::layout() {
    // show clocks in a row
    int clkx = 0;
    for (const auto &pair : clockMap) {
        Clock *clock = pair.second;
        clock->setY(-50);
        clock->setX(clkx + clock->width() / 2);
        clkx += clock->width() + 10;
    }
    // reset all levels
    for (const auto &pair : nodeMap) {
        GraphNode *node = pair.second;
        node->level(-1);
    }
    // compute levels and sizing
    int maxlevel = 0;
    for (const auto &pair : nodeMap) {
        GraphNode *node = pair.second;
        int level = node->level();
        if (level > maxlevel) {
            maxlevel = level;
        }
    }
    // move nodes to columns
    std::vector<GraphNode *> column[maxlevel + 1];
    for (const auto &pair : nodeMap) {
        GraphNode *node = pair.second;
        column[node->level()].push_back(node);
    }
    // print out columns
    int x = 0;
    for (uint16_t i = 0; i <= maxlevel; i++) {
        int y = 0;
        int maxwidth = 0;
        for (GraphNode *node : column[i]) {
            if (node->width() > maxwidth) {
                maxwidth = node->width();
            }
            node->setX(x + node->width() / 2);
            node->setY(y + node->height() / 2);
            y += node->height();
        }
        x += maxwidth + 50;
    }
}
