/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "outputwidget.h"
#include "style.h"
#include "ui_outputwidget.h"

#include <QClipboard>
#include <QDir>
#include <QKeyEvent>
#include <QMessageBox>
#include <signal.h>

void UtfChar::append(char next) {
    // do something
    if (index == 0) {
        msize = 1;
        if (next & 0x80) { // is multibyte?
            uint head = next;
            head >>= 4;
            head &= 7;
            msize = std::min((head == 4) ? 2 : head - 3, (uint)MAX_SIZE);
        }
        buffer[msize] = '\0';
    }
    buffer[index++] = next;
}

bool UtfChar::complete() { return index == msize; }

char *UtfChar::bytes() {
    index = 0;
    return buffer;
}

OutputWidget::OutputWidget(QWidget *parent)
    : QWidget(parent), mpid(0), ui(new Ui::OutputWidget), selectedItem(0) {
    ui->setupUi(this);
    mgraph = new ObjectGraph(this);
    ui->splitter2->insertWidget(0, mgraph);
    ui->splitter->setSizes({250, 750});
    ui->splitter2->setSizes({600, 150});
    errorFormat.setForeground(QColor(BIPIDE_COLOR_ERROR));
    inFormat.setForeground(QColor(BIPIDE_COLOR_INPUT));
    connect(ui->endButton, SIGNAL(clicked(bool)), this, SLOT(endScript(bool)));
    connect(ui->killButton, SIGNAL(clicked(bool)), this, SLOT(killScript(bool)));
    connect(ui->restartButton, SIGNAL(clicked(bool)), this, SLOT(restartScript(bool)));
    connect(ui->stdoutBrowser, SIGNAL(anchorClicked(const QUrl &)), this,
            SLOT(stdoutLink(const QUrl &)));
    ui->stdoutBrowser->setFocusProxy(mgraph);
    osc = new OscServer();
    connect(osc, SIGNAL(message(OscMessage)), this, SLOT(message(OscMessage)));
    connect(mgraph, SIGNAL(selected(GraphItem *)), this, SLOT(selected(GraphItem *)));
}

OutputWidget::~OutputWidget() { delete ui; }

void OutputWidget::keyPressEvent(QKeyEvent *event) {
    QString text = event->text();
    bool istext = false;
    if (event->modifiers() & Qt::ControlModifier) {
        if (event->key() == Qt::Key_V) { // Ctrl-V paste
            text = QGuiApplication::clipboard()->text();
            istext = true;
        }
    } else {
        istext = !(event->modifiers() & Qt::AltModifier); // TODO: better test for text
    }
    // emit
    emit textInput(text);
    // write to output if it's text
    if (istext) {
        QTextCursor cursor(ui->stdoutBrowser->document());
        cursor.movePosition(QTextCursor::End);
        cursor.insertText(text, inFormat);
    }
}

void OutputWidget::running(int pid) {
    QString base = tr("process ID ");
    QString mesg = base + QString::number(pid);
    ui->statusLabel->setText(mesg);
    ui->statusLabel->setStyleSheet("color: green");
    ui->endButton->setEnabled(true);
    ui->killButton->setEnabled(true);
    ui->restartButton->setEnabled(true);
    this->mpid = pid;
}

void OutputWidget::write(const char *data) {
    // position cursor to end, note position
    QTextCursor cursor(ui->stdoutBrowser->document());
    cursor.movePosition(QTextCursor::End);
    int chunkStart = cursor.position();
    // assemble utf characters
    QString output;
    for (int i = 0; data[i] != '\0'; i++) {
        mutfchar.append(data[i]);
        if (mutfchar.complete()) {
            output.append(mutfchar.bytes());
        }
    }
    cursor.insertText(output, outFormat);
    // parse output for code errors
    QString chunk(output);
    int index = chunk.indexOf("\n"); // TODO: regex
    while (index != -1) {
        lineBuffer.append(chunk.left(index));
        parseLine(chunkStart + index);
        lineBuffer.clear();
        chunk = chunk.right(chunk.length() - index - 1);
        chunkStart += index;
        index = chunk.indexOf("\n");
    }
    // append incomplete line
    lineBuffer.append(chunk);
    delete[] data;
}

void OutputWidget::writeError(QString chunk) {
    QTextCursor cursor(ui->stdoutBrowser->document());
    cursor.movePosition(QTextCursor::End);
    cursor.insertText(chunk, errorFormat);
}

void OutputWidget::parseLine(int pos) {
    // format for links
    QTextCharFormat anchorFormat;
    anchorFormat.setAnchor(true);
    anchorFormat.setForeground(QColor("#0000BB"));
    anchorFormat.setUnderlineStyle(QTextCharFormat::SingleUnderline);
    // parse error
    // [path].bip line = (2) column = (9) : error end of statement expected (; or lf)
    QRegExp perrorExp("(.*) line = \\(([0-9]+)\\) column = \\(([0-9]+)\\) : (.*)");
    if (perrorExp.indexIn(lineBuffer) > -1) {
        // make link
        QUrl href;
        href.setScheme("bipsrc");
        href.setHost("[parse]");
        href.setPort(perrorExp.cap(2).toInt());
        href.setPath(perrorExp.cap(1));
        href.setQuery(perrorExp.cap(3));
        anchorFormat.setAnchorHref(href.toString());
        // add link to document
        QTextCursor cursor(ui->stdoutBrowser->document());
        cursor.setPosition(pos);
        cursor.select(QTextCursor::BlockUnderCursor);
        cursor.setCharFormat(anchorFormat);
        // emit error
        emit codeError(href.path(), href.port());
    }
    // runtime error
    QRegExp errorExp("\\*FUNCTION \\[([^\\(]+)\\(\\)\\] (.*) line \\[([0-9]+)\\]");
    if (errorExp.indexIn(lineBuffer) > -1) {
        // make link
        QUrl href;
        href.setScheme("bipsrc");
        href.setHost(errorExp.cap(1));
        href.setPort(errorExp.cap(3).toInt());
        QString path = errorExp.cap(2);
        if (!path.startsWith("/")) {
            QFileInfo fileInfo(QDir::currentPath() + "/" + path);
            path = fileInfo.canonicalFilePath();
        }
        href.setPath(path);
        anchorFormat.setAnchorHref(href.toString());
        // add link to document
        QTextCursor cursor(ui->stdoutBrowser->document());
        cursor.setPosition(pos);
        cursor.select(QTextCursor::BlockUnderCursor);
        cursor.setCharFormat(anchorFormat);
        // emit error
        emit codeError(href.path(), href.port());
    }
}

void OutputWidget::endScript(bool) {
    if (mpid != 0) {
        kill(mpid, 15);
    }
}

void OutputWidget::killScript(bool) {
    if (mpid != 0) {
        kill(mpid, 9);
    }
}

void OutputWidget::restartScript(bool) {
    if (mpid != 0) {
        kill(mpid, 1);
    }
}
void OutputWidget::selected(GraphItem *item) {
    this->selectedItem = item;
    if (item) {
        ui->selectedLabel->setText(item->name());
        ui->connectorCombo->clear();
        ui->clockCombo->clear();
        QList<QString> cnames = item->eventConnectors();
        int index = 0;
        for (QString cname : cnames) {
            ui->connectorCombo->insertItem(index++, cname);
        }
        if (cnames.size()) {
            QList<Clock *> clocks = item->eventClocks(cnames[0]);
            for (Clock *clock : clocks) {
                ui->clockCombo->insertItem(index++, clock->name());
            }
            if (clocks.size()) {
                ui->eventTable->setModel(item->eventModel(0, 0));
                ui->eventTable->resizeColumnsToContents();
            } else {
                ui->eventTable->setModel(&EmptyTableModel::instance());
                ui->eventTable->resizeColumnsToContents();
            }
        } else {
            ui->eventTable->setModel(&EmptyTableModel::instance());
            ui->eventTable->resizeColumnsToContents();
        }
    }
}

void OutputWidget::completed(int status) {
    // disable buttons
    ui->endButton->setEnabled(false);
    ui->killButton->setEnabled(false);
    ui->restartButton->setEnabled(false);
    // show message
    QString mesg("Completed with code ");
    QString color("blue");
    int code = WEXITSTATUS(status);
    mesg.append(QString::number(code));
    if (WIFSIGNALED(status)) {
        mesg.append(" and signal ");
        mesg.append(QString::number(WTERMSIG(status)));
        color = BIPIDE_COLOR_ERROR;
    } else if (code != 0) {
        color = BIPIDE_COLOR_ERROR;
    }
    ui->statusLabel->setStyleSheet("background-color : white; color: " + color);
    ui->statusLabel->setText(mesg);
    // grey out labels
    ui->statusLabel->setEnabled(false);
    // grey out graph
    mgraph->completed();
    // set PID to zero
    mpid = 0;
}

void OutputWidget::message(OscMessage message) {
    QString path = message.path();
    int args = message.args();
    if (path == "/event/new") {
        int64_t port = message.arg(0).toLongLong();
        int64_t clock = message.arg(1).toLongLong();
        int64_t eid = message.arg(2).toLongLong();
        float measure = message.arg(3).toFloat();
        QString name = message.arg(4).toString();
        int16_t arg1 = message.arg(5).toInt();
        int16_t arg2 = message.arg(6).toInt();
        mgraph->addMidiEvent(port, clock, eid, measure, name, arg1, arg2);
    } else if (path == "/event/remove") {
        int64_t port = message.arg(0).toLongLong();
        int64_t clock = message.arg(1).toLongLong();
        int64_t eid = message.arg(2).toLongLong();
        mgraph->removeMidiEvent(port, clock, eid);
    } else if (path == "/event/clear") {
        int64_t port = message.arg(0).toLongLong();
        int64_t clock = message.arg(1).toLongLong();
        mgraph->clearMidiEvents(port, clock);
    } else if (path == "/event/script/new") {
        int64_t clock = message.arg(0).toLongLong();
        int64_t eid = message.arg(1).toLongLong();
        float measure = message.arg(2).toFloat();
        QString name = message.arg(3).toString();
        mgraph->addScriptEvent(clock, eid, measure, name);
    } else if (path == "/event/script/remove") {
        int64_t clock = message.arg(0).toLongLong();
        int64_t eid = message.arg(1).toLongLong();
        mgraph->removeScriptEvent(clock, eid);
    } else if (path == "/event/script/clear") {
        int64_t clock = message.arg(0).toLongLong();
        mgraph->clearScriptEvents(clock);
    } else if (path == "/node/level") {
        int64_t node = message.arg(0).toLongLong();
        int levels = args - 1;
        if (levels) {
            float level[levels];
            for (int i = 0; i < levels; i++) {
                level[i] = message.arg(1 + i).toFloat();
            }
            mgraph->levels(node, level);
        }
    } else if (path == "/node/midilevel" && args == 3) {
        int64_t node = message.arg(0).toLongLong();
        int16_t output = message.arg(1).toInt();
        bool activity = message.arg(2).toBool();
        mgraph->midilevels(node, output, activity);
    } else if (path == "/clock/new" && (args == 3 || args == 5)) {
        int64_t node = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        float tempo = message.arg(2).toFloat();
        if (args == 5) {
            uint16_t signum = message.arg(3).toInt();
            uint16_t sigden = message.arg(4).toInt();
            mgraph->addClock(node, name, tempo, signum, sigden);
        } else {
            mgraph->addClock(node, name, tempo, 4, 4);
        }
    } else if (path == "/clock/time" && args == 2) {
        int64_t node = message.arg(0).toLongLong();
        float measure = message.arg(1).toFloat();
        mgraph->clockTime(node, measure);
    } else if (path == "/clock/update" && args == 6) {
        int64_t node = message.arg(0).toLongLong();
        bool running = message.arg(1).toBool();
        float measure = message.arg(2).toFloat();
        float tempo = message.arg(3).toFloat();
        uint16_t signum = message.arg(4).toInt();
        uint16_t sigden = message.arg(5).toInt();
        mgraph->clockUpdate(node, running, measure, tempo, signum, sigden);
    } else if (path == "/clock/finish" && args == 1) {
        int64_t node = message.arg(0).toLongLong();
        mgraph->finishClock(node);
    } else if (path == "/connect" && args == 2) {
        int64_t port1 = message.arg(0).toLongLong();
        int64_t port2 = message.arg(1).toLongLong();
        mgraph->cconnect(port1, port2);
    } else if (path == "/lv2/new" && args == 6) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addLv2Plugin(hash, name, message.arg(2).toInt(), message.arg(3).toInt(),
                             message.arg(4).toInt(), message.arg(5).toInt());
    } else if (path == "/mixer/new" && args == 4) {
        int64_t mixer = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addMixer(mixer, name, message.arg(2).toInt(), message.arg(3).toInt());
    } else if (path == "/input/new" && args == 4) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addInput(hash, name, message.arg(3).toInt());
    } else if (path == "/output/new" && args == 4) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addOutput(hash, name, message.arg(2).toInt());
    } else if (path == "/midiin/new" && args == 4) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addMidiInput(hash, name, message.arg(3).toInt());
    } else if (path == "/midiout/new" && args == 4) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addMidiOutput(hash, name, message.arg(2).toInt());
    } else if (path == "/processor/new" && args == 6) {
        int64_t hash = message.arg(0).toLongLong();
        QString name = message.arg(1).toString();
        mgraph->addLv2Plugin(hash, name, message.arg(2).toInt(), message.arg(3).toInt(),
                             message.arg(4).toInt(), message.arg(5).toInt());
    } else if (path == "/processor/addmidi" && args == 3) {
        int64_t hash = message.arg(0).toLongLong();
        mgraph->addProcessorMidi(hash, message.arg(1).toInt(), message.arg(2).toInt());
    } else if (path == "/node/connectors") {
        int64_t node = message.arg(0).toLongLong();
        bool ismidi = message.arg(1).toBool();
        int inputs = args - 2;
        if (inputs) {
            int64_t port[inputs];
            for (int i = 0; i < inputs; i++) {
                port[i] = message.arg(2 + i).toLongLong();
            }
            if (ismidi) {
                mgraph->nodeMidiInputs(node, port);
            } else {
                mgraph->nodeInputs(node, port);
            }
        }
    } else if (path == "/node/connections") {
        int64_t node = message.arg(0).toLongLong();
        bool ismidi = message.arg(1).toBool();
        int outputs = args - 2;
        int64_t port[outputs];
        for (int i = 0; i < outputs; i++) {
            port[i] = message.arg(2 + i).toLongLong();
        }
        if (ismidi) {
            mgraph->nodeMidiOutputs(node, port);
        } else {
            mgraph->nodeOutputs(node, port);
        }
    } else if (path == "/node/remove" && args == 1) {
        int64_t hash = message.arg(0).toLongLong();
        mgraph->removeNode(hash);
    } else if (path == "/mixer/gain") {
        int64_t mixer = message.arg(0).toLongLong();
        int input = message.arg(1).toInt();
        int output = message.arg(2).toInt();
        float amount = message.arg(3).toFloat();
        mgraph->mixerGain(mixer, input, output, amount);
    } else if (path == "/mixer/gains") {
        int64_t mixer = message.arg(0).toLongLong();
        int input = message.arg(1).toInt();
        int outputs = args - 2;
        float gains[outputs];
        for (int i = 0; i < outputs; i++) {
            gains[i] = message.arg(2 + i).toFloat();
        }
        mgraph->mixerGains(mixer, input, gains);
    } else if (path == "/script/start") {
        mgraph->scriptStart();
    } else {
        // TODO: show error
        printf("Unknown path or wrong number of arguments (%d): %s\n", args,
               path.toStdString().c_str());
    }
}

void OutputWidget::updateModel(int connectoridx, int clockidx) {
    if (selectedItem && connectoridx >= 0 && clockidx >= 0) {
        ui->eventTable->setModel(selectedItem->eventModel(connectoridx, clockidx));
        ui->eventTable->resizeColumnsToContents();
    }
}

void OutputWidget::on_connectorCombo_currentIndexChanged(int connectoridx) {
    int clockidx = ui->clockCombo->currentIndex();
    updateModel(connectoridx, clockidx);
}

void OutputWidget::on_clockCombo_currentIndexChanged(int clockidx) {
    int connectoridx = ui->connectorCombo->currentIndex();
    updateModel(connectoridx, clockidx);
}
