/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "projectconfig.h"
#include "ext/RSJp-cpp/RSJparser.tcc"
#include <QDebug>
#include <QDirIterator>
#include <QStandardPaths>
#include <QTextStream>
#include <algorithm>
#include <fstream>

#define BIP_IDE_CONFIG_EXTENSION ".json"

QString ProjectConfig::filename(QString id) {
    QDir configdir = getConfigDir();
    return configdir.canonicalPath() + "/" + id + BIP_IDE_CONFIG_EXTENSION;
}

Project *ProjectConfig::findByLocation(QString folder) {
    QDir dir(folder);
    QFileInfo folderinfo(dir.canonicalPath());
    for (ProjectId name : names()) {
        Project *project = load(name.id);
        for (QString path : project->folders()) {
            QFileInfo info(path);
            if (info == folderinfo) {
                return project;
            }
        }
    }
    return nullptr;
}

QString ProjectConfig::createId() { return QUuid::createUuid().toString(QUuid::WithoutBraces); }

bool ProjectConfig::save(Project &project) {
    QFile file(filename(project.id()));
    if (!file.open(QIODevice::WriteOnly)) {
        return false;
    }
    QTextStream stream(&file);
    stream << "{\"name\": \"" << project.name() << "\"";
    stream << ",\"exec\": \"" << project.executable() << "\"";
    stream << ",\"active\": \"" << project.activeFile() << "\"";
    // folders
    stream << ",\"folders\": [";
    QList<QString> folders = project.folders();
    for (int i = 0; i < folders.size(); i++) {
        if (i) {
            stream << ",";
        }
        stream << "\"" << folders.at(i) << "\"";
    }
    stream << "]";
    // open files
    stream << ",\"open\": [";
    QList<QString> open = project.openFiles();
    for (int i = 0; i < open.size(); i++) {
        if (i) {
            stream << ",";
        }
        stream << "\"" << open.at(i) << "\"";
    }
    stream << "]";
    // run configs
    stream << ",\"run\": {\"active\" : " << project.activeRun();
    stream << ", \"config\" : [";
    const QList<RunConfig> &runs = project.runConfigs();
    for (int i = 0; i < runs.size(); i++) {
        if (i) {
            stream << ",";
        }
        stream << "{\"path\": \"" << runs.at(i).fullpath() << "\",";
        stream << "\"name\": \"" << runs.at(i).name() << "\",";
        stream << "\"args\": \"" << runs.at(i).args() << "\"}";
    }
    stream << "]}";
    stream << "}";
    file.close();
    return true;
}

bool ProjectConfig::remove(QString id) {
    QFile file(filename(id));
    return file.remove();
}

Project *ProjectConfig::load(QString id) {
    QFileInfo file(filename(id));
    if (!file.exists() || !file.isReadable()) {
        return nullptr;
    }
    std::ifstream fs(filename(id).toStdString());
    RSJresource json(fs);
    Project *project = new Project(id);
    // name
    std::string name = json["name"].as<std::string>("[project name]");
    project->name(QString::fromStdString(name));
    // executable
    std::string exec = json["exec"].as<std::string>("/usr/bin/bipscript");
    project->executable(QString::fromStdString(exec));
    // active file
    std::string active = json["active"].as<std::string>("");
    project->activeFile(QString::fromStdString(active));
    // folders
    RSJarray folders = json["folders"].as_array();
    for (size_t i = 0; i < folders.size(); i++) {
        project->folders().append(QString::fromStdString(folders[i].as<std::string>()));
    }
    // open files
    RSJarray open = json["open"].as_array();
    for (size_t i = 0; i < open.size(); i++) {
        project->openFiles().append(QString::fromStdString(open[i].as<std::string>()));
    }
    // run configs
    project->activeRun(json["run"]["active"].as<int>(0));
    RSJarray runs = json["run"]["config"].as_array();
    for (size_t i = 0; i < runs.size(); i++) {
        QString name = QString::fromStdString(runs[i]["name"].as<std::string>());
        QString path = QString::fromStdString(runs[i]["path"].as<std::string>());
        QString args = QString::fromStdString(runs[i]["args"].as<std::string>());
        if (name.length()) {
            RunConfig run(name, path, args);
            project->runConfigs().append(run);
        } else {
            RunConfig run(path, args);
            project->runConfigs().append(run);
        }
    }
    fs.close();
    return project;
}

int ProjectConfig::count() { // need this?
    int ret = 0;
    QDirIterator it(getConfigDir().path(), QStringList() << "*" BIP_IDE_CONFIG_EXTENSION,
                    QDir::Files);
    while (it.hasNext()) {
        it.next();
        ret++;
    }
    return ret;
}

QList<ProjectId> ProjectConfig::names() {
    QList<ProjectId> ret;
    QDirIterator it(getConfigDir().path(), QStringList() << "*" BIP_IDE_CONFIG_EXTENSION,
                    QDir::Files);
    while (it.hasNext()) {
        std::string filename = it.next().toStdString();
        QFileInfo fileinfo(filename.c_str());
        QString id = fileinfo.baseName();
        std::ifstream file(filename);
        RSJresource json(file);
        std::string name = json["name"].as<std::string>("");
        ProjectId idobj;
        idobj.name = QString::fromStdString(name);
        idobj.id = id;
        ret.append(idobj);
        file.close();
    }
    std::sort(ret.begin(), ret.end());
    return ret;
}

QDir ProjectConfig::getConfigDir() {
    QString base = QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
    QDir ret = QDir(base + "/" + BIP_IDE_CONFIG_DIR);
    if (!ret.exists()) {
        ret.mkpath(ret.path());
    }
    return ret;
}
