/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "projectlist.h"

#include <QDir>
#include <QSettings>
#include <QUuid>

LegacyProject::LegacyProject(QString id) : mid(id) { this->executable("/usr/bin/bipscript"); }

bool LegacyProject::exists() {
    for (int i = 0; i < folders.size(); i++) {
        QDir dir(folders[i].path);
        if (!dir.exists()) {
            return false;
        }
    }
    return true;
}

void ProjectList::load() {
    QSettings settings;
    int size = settings.beginReadArray("projects");
    for (int i = 0; i < size; ++i) {
        settings.setArrayIndex(i);
        LegacyProject project(QUuid::createUuid().toString(QUuid::WithoutBraces));
        project.executable(settings.value("exec").toString());
        // folders
        int fsize = settings.beginReadArray("folders");
        for (int j = 0; j < fsize; j++) {
            settings.setArrayIndex(j);
            NamedFile folder;
            folder.name = settings.value("name").toString();
            folder.path = settings.value("path").toString();
            project.addFolder(folder);
        }
        settings.endArray();
        // files
        fsize = settings.beginReadArray("open");
        for (int j = 0; j < fsize; j++) {
            settings.setArrayIndex(j);
            project.openFile(settings.value("path").toString());
        }
        settings.endArray();
        project.activeFile(settings.value("active").toString());
        projects.append(project);
    }
    settings.endArray();
    openIndex = settings.value("appstate/project").toInt();
    if (openIndex >= projects.size()) {
        openIndex = projects.size() - 1;
    }
}

void ProjectList::save() {
    QSettings settings;
    settings.beginWriteArray("projects");
    for (int i = 0; i < projects.size(); i++) {
        settings.setArrayIndex(i);
        settings.setValue("exec", projects[i].executable());
        // folders
        settings.beginWriteArray("folders");
        for (int j = 0; j < projects[i].folderCount(); j++) {
            settings.setArrayIndex(j);
            NamedFile &folder = projects[i].folder(j);
            settings.setValue("name", folder.name);
            settings.setValue("path", folder.path);
        }
        settings.endArray();
        // files
        int count = 0;
        settings.beginWriteArray("open");
        foreach (const QString &path, projects[i].openFiles()) {
            settings.setArrayIndex(count++);
            settings.setValue("path", path);
        }
        settings.endArray();
        settings.setValue("active", projects[i].activeFile());
    }
    settings.endArray();
    settings.setValue("appstate/project", openIndex);
}

bool ProjectList::currentExists() { return size() > 0 && currentProject().exists(); }

QString ProjectList::commonRoot() {
    QString root = get(0).folder(0).path;
    for (int i = 0; i < size(); i++) {
        LegacyProject &project = get(i);
        for (int j = 0; j < project.folderCount(); j++) {
            root = commonRoot(root, project.folder(j).path);
        }
    }
    return root;
}

// TODO: generalized method to match multiple paths
int ProjectList::findProject(QString &path) {
    for (int i = 0; i < size(); i++) {
        LegacyProject &project = get(i);
        if (project.folderCount() == 1) {
            NamedFile folder = project.folder(0);
            if (folder.path == path) {
                return i;
            }
        }
    }
    return -1;
}

void ProjectList::removeProject(int index) {
    if (index < openIndex) {
        openIndex--;
    }
    projects.removeAt(index);
}

bool ProjectList::open(int index) {
    // check if exists
    if (get(index).exists()) {
        openIndex = index;
        return true;
    }
    return false;
}

bool ProjectList::select(QString path) {
    int index = findProject(path);
    return index == -1 ? false : open(index);
}

QString ProjectList::commonRoot(QString &path1, QString &path2) {
    QStringList list1 = path1.split(QDir::separator());
    QStringList list2 = path2.split(QDir::separator());
    QString ret;
    int i = 0;
    while (i < list1.size() && i < list2.size() && list1[i] == list2[i]) {
        if (list1[i].size() > 0) {
            ret.append(QDir::separator());
            ret.append(list1[i]);
        }
        i++;
    }
    return ret;
}
