/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "projectsdialog.h"
#include "ui_projectsdialog.h"

#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>

ProjectsDialog::ProjectsDialog(QWidget *parent, Project *current)
    : QDialog(parent), mcurrent(current), ui(new Ui::ProjectsDialog) {
    ui->setupUi(this);
    // populate list
    syncList();
}

ProjectsDialog::~ProjectsDialog() { delete ui; }

void ProjectsDialog::on_cancelButton_clicked() { this->close(); }

void ProjectsDialog::syncList() {
    ui->listWidget->clear();
    ui->openButton->setEnabled(false);
    ui->removeButton->setEnabled(false);
    QList<ProjectId> names = config.names();
    for (auto project : names) {
        QListWidgetItem *item = new QListWidgetItem;
        item->setText(project.name);
        item->setData(Qt::UserRole, project.id);
        ui->listWidget->addItem(item);
    }
}

void ProjectsDialog::on_newButton_clicked() {
    QString root(QDir::homePath());
    QString path = QFileDialog::getExistingDirectory(this, "Open Project Folder", root);
    if (path.length() > 0) {
        // need new
        QFileInfo fileInfo(path);
        NamedFile folder(fileInfo.fileName(), fileInfo.filePath());
        QString message = "Please give this project a name";
        bool ok;
        QString name =
            QInputDialog::getText(this, "Project Name", message, QLineEdit::Normal, "", &ok);
        Project project(config.createId());
        project.name(name);
        project.folders().append(fileInfo.filePath());
        if (config.save(project)) {
            syncList();
            // select the new one
            auto items = ui->listWidget->findItems(name, Qt::MatchExactly);
            if (items.size()) {
                ui->listWidget->setCurrentItem(items.at(0));
            }
        } else {
            QString message = QString("Could not save project file to " BIP_IDE_CONFIG_DIR);
            QMessageBox::warning(this, tr("Error Saving File"), message, QMessageBox::Ok);
        }
    }
}

void ProjectsDialog::on_openButton_clicked() {
    QListWidgetItem *item = ui->listWidget->currentItem();
    QString pid = item->data(Qt::UserRole).toString();
    if (config.load(pid)) {
        emit projectChanged(pid);
        this->close();
    } else {
        QString message("Cannot open project file at " + config.filename(pid));
        QMessageBox::warning(this, "Project Config", message);
    }
}

void ProjectsDialog::on_removeButton_clicked() {
    QMessageBox msgBox(this);
    msgBox.setText("Remove Project");
    msgBox.setInformativeText("Do you want to remove this project?");
    msgBox.setStandardButtons(QMessageBox::Ok | QMessageBox::Cancel);
    msgBox.setDefaultButton(QMessageBox::Cancel);
    int ret = msgBox.exec();
    if (ret == QMessageBox::Ok) {
        QListWidgetItem *item = ui->listWidget->currentItem();
        QString pid = item->data(Qt::UserRole).toString();
        config.remove(pid);
        syncList();
    }
}

void ProjectsDialog::on_listWidget_itemSelectionChanged() {
    int selection = ui->listWidget->currentRow();
    bool selected(selection != -1);
    ui->openButton->setEnabled(selected);
    // disable remove button for open project
    bool enabled = false;
    if (selected) {
        QListWidgetItem *item = ui->listWidget->currentItem();
        QString pid = item->data(Qt::UserRole).toString();
        enabled = !mcurrent || pid != mcurrent->id();
    }
    ui->editButton->setEnabled(enabled);
    ui->openButton->setEnabled(enabled);
    ui->removeButton->setEnabled(enabled);
}

void ProjectsDialog::on_editButton_clicked() {
    QListWidgetItem *item = ui->listWidget->currentItem();
    QString pid = item->data(Qt::UserRole).toString();
    Project *project = config.load(pid);
    if (project) {
        bool ok;
        QString message = "Rename project:";
        QString name = QInputDialog::getText(this, "Project Name", message, QLineEdit::Normal,
                                             project->name(), &ok);
        if (ok) {
            project->name(name);
            config.save(*project);
            syncList();
        }
    } else {
        QString message("Cannot edit project file at " + config.filename(pid));
        QMessageBox::warning(this, "Project Config", message);
    }
}
