/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "scriptrun.h"

#include <poll.h>
#include <spawn.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>

QString ScriptRun::executable;

extern char **environ;

void ScriptRun::setExecutable(QString &exec) { executable = exec; }

bool ScriptRun::possible() {
    // does executable exist?
    return access(executable.toStdString().c_str(), X_OK) != -1;
}

void ScriptRun::writeStdin(QString input) {
    std::string string = input.toStdString().c_str();
    size_t len = string.size();
    char *buffer = new char[len + 1];
    strncpy(buffer, string.c_str(), len);
    buffer[len] = '\0';
    if (!inputq.push(buffer)) {
        // TODO: show warning: typed input was lost
        delete[] buffer;
    }
}

void ScriptRun::run() {
    chdir(script.path().toStdString().c_str());

    // set up command line
    pid_t pid;
    char *filepath = strdup(script.filePath().toStdString().c_str());
    char *exec = strdup(executable.toStdString().c_str());
    char *ounbuffer = strdup("--unbuffer-stdout");
    char *ostartport = strdup("--status-port");
    char *osighub = strdup("--sighup-restart");
    char portnum[64];
    sprintf(portnum, "%d", mport);

    // parse command line args
    QStringList userargs;
    if (marguments.size()) {
        QRegExp exp("(\"[^\"]*\")|([^\\s]+)");
        int index = 0;
        while ((index = exp.indexIn(marguments, index)) != -1) {
            userargs.append(exp.cap(0));
            index += exp.matchedLength();
        }
    }

    int argc = userargs.size();
    char **argv = new char *[7 + argc];
    argv[0] = exec;
    argv[1] = ounbuffer;
    argv[2] = ostartport;
    argv[3] = portnum;
    argv[4] = osighub;
    argv[5] = filepath;
    for (int i = 0; i < argc; i++) {
        argv[6 + i] = strdup(userargs[i].toStdString().c_str());
    }
    argv[6 + argc] = nullptr;

    // set up output redirection
    int cout_pipe[2];
    int cerr_pipe[2];
    int cin_pipe[2];
    posix_spawn_file_actions_t actions;
    posix_spawn_file_actions_init(&actions);
    if (pipe(cout_pipe)) {
        emit error("Failed to create stdout pipes");
    } else {
        posix_spawn_file_actions_addclose(&actions, cout_pipe[0]);
        posix_spawn_file_actions_adddup2(&actions, cout_pipe[1], STDOUT_FILENO);
        posix_spawn_file_actions_addclose(&actions, cout_pipe[1]);
    }
    if (pipe(cerr_pipe)) {
        emit error("Failed to create stderr pipes");
    } else {
        posix_spawn_file_actions_addclose(&actions, cerr_pipe[0]);
        posix_spawn_file_actions_adddup2(&actions, cerr_pipe[1], STDERR_FILENO);
        posix_spawn_file_actions_addclose(&actions, cerr_pipe[1]);
    }
    if (pipe(cin_pipe)) {
        emit error("Failed to create stdin pipes");
    } else {
        posix_spawn_file_actions_addclose(&actions, cin_pipe[1]);
        posix_spawn_file_actions_adddup2(&actions, cin_pipe[0], STDIN_FILENO);
        posix_spawn_file_actions_addclose(&actions, cin_pipe[0]);
    }

    int status = posix_spawn(&pid, exec, &actions, nullptr, argv, environ);

    delete[] argv; // TODO: delete items too

    close(cout_pipe[1]);
    close(cerr_pipe[1]);
    close(cin_pipe[0]);

    if (status == 0) {
        // emit started
        emit started(pid);

        // buffer for stdout + stderr
        static const int BUFFER_LEN = 4096;
        char buffer[BUFFER_LEN];
        pollfd pollfds[] = {{cout_pipe[0], POLLIN, 0}, {cerr_pipe[0], POLLIN, 0}};

        // main input/output loop
        while (true) {

            // poll stdout stderr for activity and emit as string
            int result = poll(pollfds, 2, 0);
            if (pollfds[0].revents & POLLIN) {
                long len = read(cout_pipe[0], &buffer, BUFFER_LEN - 1);
                buffer[len] = '\0';
                char *copy = new char[len + 1]; // deleted in OutputWidget::write
                strncpy(copy, buffer, len + 1);
                emit readStdout(copy);
            } else if (pollfds[1].revents & POLLIN) {
                long len = read(cerr_pipe[0], &buffer, BUFFER_LEN - 1);
                buffer[len] = '\0';
                emit readStderr(QString(buffer));
            } else if (result > 0) {
                break;
            }

            // pull in input from slot thread for stdin
            const char *input;
            while (inputq.pop(input)) {
                write(cin_pipe[1], input, strlen(input));
                delete[] input;
            }
            usleep(1000);
        }
        // wait and emit exit code
        int status;
        waitpid(pid, &status, 0);
        posix_spawn_file_actions_destroy(&actions);
        emit exited(status);
    } else {
        emit error(strerror(status));
    }
    this->deleteLater();
}
