/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "searchwidget.h"
#include "ui_searchwidget.h"

SearchWidget::SearchWidget(QWidget *parent)
    : QWidget(parent), ui(new Ui::SearchWidget), mnewSearch(true), count(0) {
    ui->setupUi(this);
    connect(ui->treeWidget, SIGNAL(itemClicked(QTreeWidgetItem *, int)), this,
            SLOT(resultTreeClicked(QTreeWidgetItem *, int)));
}

void SearchWidget::addResults(const QString &path, const QString &filename,
                              QList<EditorLink> results, bool open) {
    // header tree node for file name
    SearchTreeFileItem *parent = new SearchTreeFileItem(path, filename, results.size());
    ui->treeWidget->addTopLevelItem(parent);
    if (open) {
        parent->setExpanded(true);
    }
    // child tree nodes for results
    foreach (EditorLink link, results) {
        QTreeWidgetItem *item = new SearchTreeItem(parent, link);
        parent->addChild(item);
    }
    // update counts
    count += results.size();
    fcount++;
    showCount();
}

void SearchWidget::updateResults(const QString &path, QList<EditorLink> results) {
    QTreeWidgetItem *root = ui->treeWidget->invisibleRootItem();
    // child tree nodes for this path
    for (int i = 0; i < root->childCount(); i++) {
        SearchTreeFileItem *item = static_cast<SearchTreeFileItem *>(root->child(i));
        if (item->path() == path) {
            count += results.size() - item->results();
            item->takeChildren();
            foreach (EditorLink link, results) {
                QTreeWidgetItem *child = new SearchTreeItem(item, link);
                item->addChild(child);
            }
            item->results(results.size());
            showCount();
        }
    }
}

void SearchWidget::clear() {
    ui->searchLineEdit->clear();
    ui->treeWidget->clear();
    ui->countLabel->clear();
    ui->fileScopeButton->setChecked(true);
}

SearchWidget::~SearchWidget() { delete ui; }

void SearchWidget::resultTreeClicked(QTreeWidgetItem *treeItem, int) {
    SearchTreeItem *item = dynamic_cast<SearchTreeItem *>(treeItem);
    if (item) {
        emit navigateResult(item->location());
    }
}

void SearchWidget::on_searchButton_clicked() { doSearch(); }

void SearchWidget::on_searchLineEdit_returnPressed() { doSearch(); }

void SearchWidget::on_searchLineEdit_textEdited(const QString &) { mnewSearch = true; }

void SearchWidget::on_replaceButton_clicked() {
    QLineEdit *lineEdit = this->ui->searchLineEdit;
    bool sensitive = ui->searchSensistiveCheck->isChecked();
    SearchTerm term(lineEdit->text(), sensitive);
    QString replace = this->ui->replaceLineEdit->text();
    emit replaceNext(term, replace);
}

void SearchWidget::on_refreshButton_clicked() {
    mnewSearch = true;
    doSearch();
}

void SearchWidget::doSearch() {
    QLineEdit *lineEdit = this->ui->searchLineEdit;
    bool sensitive = ui->searchSensistiveCheck->isChecked();
    SearchTerm search(lineEdit->text(), sensitive);

    if (mnewSearch) {
        count = 0;
        fcount = 0;
        showCount();
        ui->treeWidget->clear();
        if (search.string().size() == 0) {
            emit clearSearch();
        } else if (ui->fileScopeButton->isChecked()) {
            emit searchCurrent(search);
            mnewSearch = false;
        } else if (ui->projectScopeButton->isChecked()) {
            emit searchProject(search);
            mnewSearch = false;
        }
    } else {
        ui->treeWidget->clearSelection();
        emit searchNext(search);
    }
}

void SearchWidget::newSearch(QString term, bool project) {
    ui->searchLineEdit->setText(term);
    ui->searchLineEdit->setFocus();
    ui->searchLineEdit->selectAll();
    ui->projectScopeButton->setChecked(project);
    mnewSearch = true;
}

void SearchWidget::showCount() {
    QString label(count == 1 ? "result" : "results");
    QString flabel(fcount == 1 ? "file" : "files");
    QString mesg = QString("%1 %2 in %3 %4").arg(count).arg(label).arg(fcount).arg(flabel);
    ui->countLabel->setText(mesg);
}

void SearchWidget::on_fileScopeButton_clicked() { mnewSearch = true; }

void SearchWidget::on_projectScopeButton_clicked() { mnewSearch = true; }

void SearchWidget::on_searchSensistiveCheck_stateChanged(int) { mnewSearch = true; }
