/*
 * This file is part of Bipscript-IDE.
 *
 * Bipscript-IDE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript-IDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript-IDE.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "synhighlighter.h"

#include <QtGui>

SynHighlighter::SynHighlighter(QTextDocument *parent)
    : QSyntaxHighlighter(parent), startString("@\""), endString("\""), startComment("/*"),
      endComment("*/") {
    stringFormat.setForeground(Qt::blue);
    commentFormat.setForeground(QColor("#2B8204"));

    QTextCharFormat format;

    // plain words + numbers
    format.setForeground(QColor("#092E65"));
    addRule(QRegExp("\\b[A-Za-z0-9_]+\\b"), format);

    // function call
    format.setForeground(QColor("#037B9B"));
    addRule(QRegExp("\\b[A-Za-z0-9_]+(?=\\()"), format);

    // CamelCase
    format.setForeground(QColor("#950A75"));
    addRule(QRegExp("\\b[A-Z]+[A-Za-z0-9_]+\\b"), format);

    // keywords
    format.setForeground(QColor("#958515"));
    QStringList keywords = {
        "base",    "break",  "case",   "catch",    "class",   "clone",  "continue",    "const",
        "default", "delete", "else",   "enum",     "extends", "for",    "foreach",     "function",
        "if",      "import", "in",     "local",    "null",    "resume", "return",      "switch",
        "this",    "throw",  "try",    "typeof",   "while",   "yield",  "constructor", "instanceof",
        "true",    "false",  "static", "__LINE__", "__FILE__"};
    foreach (const QString &word, keywords) {
        addRule(QRegExp("\\b" + word + "\\b"), format);
    }

    // 1 line string
    addRule(QRegExp("\"[^\"]*\""), stringFormat);

    // 1 line comment
    addRule(QRegExp("//[^\n]*"), commentFormat);
}

void SynHighlighter::addRule(const QRegExp &regex, const QTextCharFormat &format) {
    HighlighterRule rule;
    rule.pattern = regex;
    rule.format = format;
    rules.append(rule);
}

void SynHighlighter::highlightBlock(const QString &text) {
    foreach (const HighlighterRule &rule, rules) {
        QRegExp expression(rule.pattern);
        int index = expression.indexIn(text);
        while (index >= 0) {
            int length = expression.matchedLength();
            setFormat(index, length, rule.format);
            index = expression.indexIn(text, index + length);
        }
    }
    setCurrentBlockState(0);

    int stringStart = 0;
    int searchStart = 0;
    if (previousBlockState() != 1) {
        stringStart = text.indexOf(startString);
        searchStart = stringStart + startString.length();
    }
    while (stringStart >= 0) {
        int stringEnd = text.indexOf(endString, searchStart);
        if (stringEnd == -1) {
            setCurrentBlockState(1);
            stringEnd = text.length();
        } else {
            setCurrentBlockState(0);
            stringEnd += endString.length();
        }
        setFormat(stringStart, stringEnd - stringStart, stringFormat);
        stringStart = text.indexOf(startString, stringEnd);
        searchStart = stringStart + startString.length();
    }
    int currentState = currentBlockState();

    int commentStart = 0;
    searchStart = 0;
    if (previousBlockState() != 2) {
        commentStart = text.indexOf(startComment);
        searchStart = commentStart + startComment.length();
    }
    while (commentStart >= 0) {
        int commentEnd = text.indexOf(endComment, searchStart);
        if (commentEnd == -1) {
            setCurrentBlockState(2);
            commentEnd = text.length();
        } else {
            setCurrentBlockState(currentState);
            commentEnd += endComment.length();
        }
        setFormat(commentStart, commentEnd - commentStart, commentFormat);
        commentStart = text.indexOf(startComment, commentEnd);
        searchStart = commentStart + startComment.length();
    }
}
